#' DE Analysis Server Component
#'  
#' This server module handles the server-side logic for de analysis using DESeq2
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the data with matrix, metadata and counts filter
#'         
#' @export
#'
deanalysisServer <- function(id, dataObject) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    DE_results = reactiveVal(NULL)
    data = reactiveVal(NULL)
    shinyjs::hide("downloadde_table")
    shinyjs::hide("download_condition_table_manual")
    
    observe({
      req(dataObject())
      dds = dataObject()[[1]]
      updateSelectizeInput(session, "controls", choices = colnames(dds))
    })
    
    observe({
      req(dataObject())
      dds = dataObject()[[1]]
      updateSelectizeInput(session, "treatments", choices = colnames(dds))
    })
    
    observe({
      req(dataObject(), input$sample_select)
      if(input$sample_select == 'Perform DE based on condition column in metadata') {
        condition_column = dataObject()[[4]]
        dds = dataObject()[[1]]
        if(!is.null(condition_column)) {
          metadata_show = data.frame(colData(dds))
          if(("Samples_Found" %in% colnames(metadata_show)) && (condition_column == "conditionAddedbyCogentDS")) {
            shinyjs::hide("controls_set")
            shinyjs::hide("treatment_set")
            shinyjs::hide("fdr_filter_cond")
            shinyjs::hide("logFC_filter_cond")
            shinyalert("conditional column in metadata is not found", "Please run analysis selecting samples manually", type = "warning", confirmButtonCol = "#1c82e0")
          } else {
            metadata_show_final = metadata_show %>% dplyr::select(condition_column)
            metadata_show_final = data.frame(metadata_show_final)
            output$condition_table = renderDataTable({
              DT::datatable(metadata_show_final, fillContainer = TRUE)
            })
          }
        } else {
          shinyalert("conditional column not found", "Fatal Error check data upload", type = "error", confirmButtonCol = "#1c82e0")
        }
      }
    })

    observe({
      req(dataObject())
      condition_column = dataObject()[[4]]
      dds = dataObject()[[1]]
      if((!is.null(condition_column)) && (condition_column != "Samples_Found")) {
        metadata_show = data.frame(colData(dds))
        metadata_show = metadata_show %>% dplyr::select(condition_column)
        idents = unique(metadata_show[[condition_column]])
        updateSelectizeInput(session, "controls_set", choices = idents, selected = '')
      }
    })
    
    observe({
      req(dataObject())
      condition_column = dataObject()[[4]]
      dds = dataObject()[[1]]
      if((!is.null(condition_column)) && (condition_column != "Samples_Found")) {
        metadata_show = data.frame(colData(dds))
        metadata_show = metadata_show %>% dplyr::select(condition_column)
        idents = unique(metadata_show[[condition_column]])
        updateSelectizeInput(session, "treatment_set", choices = idents, selected = '')
      }
    })
    
    DE_results <- eventReactive(input$performde, {
      
      # Check if controls and treatments are selected and valid
      if(input$sample_select == 'Select Samples Manually for comparison' && input$performde > 0) {
        if((is.null(input$controls)) || (is.null(input$treatments)) || (length(intersect(input$controls, input$treatments)) >= 1)) {
          shinyalert("Input Error", "Please select valid controls and treatments group for cosmparison. Controls and treatments should have different selections", type = "error", confirmButtonCol = "#1c82e0")
          return(NULL)
        }
      }

      # Check if controls and treatments are selected and valid
      if(input$sample_select == 'Select Samples Manually for comparison' && input$performde > 0) {
        if((length(input$controls) == 1) && (length(input$treatments) == 1)) {
          shinyalert("Input Error", "Please see Currently Differential Expression needs more than 1 sample in control or treatment", type = "error", confirmButtonCol = "#1c82e0")
          return(NULL)
        }
      }

      # Check if controls and treatments are selected and valid from metadata
      if(input$sample_select == 'Perform DE based on condition column in metadata' && input$performde > 0) {
        if((is.null(input$controls_set)) || (is.null(input$treatment_set)) || (length(intersect(input$controls_set, input$treatment_set)) >= 1)) {
          shinyalert("Input Error", "Please select valid controls and treatments group for comparison", type = "error", confirmButtonCol = "#1c82e0")
          return(NULL)
        }
      }
      
      req(dataObject())
      dds = dataObject()[[1]]
      metadata = dataObject()[[2]]
      countsfilter = dataObject()[[3]]
      condition_column = dataObject()[[4]]
      if(input$sample_select == 'Select Samples Manually for comparison') {
        req(input$controls, input$treatments)
        withProgress(message = 'Performing DE Analysis...', value = 0, {
          incProgress(0.1, detail = "Checking Data and parameters...")
          tryCatch({
            if(!is.null(input$controls) & !is.null(input$treatments)) {
              control_set_use = rep("Control", length(input$controls))
              treatment_set_use = rep("Treatment", length(input$treatments))
              controls_df = cbind(control_set_use, input$controls)
              colnames(controls_df) = c('condition', 'Samples')
              treatment_df = cbind(treatment_set_use, input$treatments)
              colnames(treatment_df) = c('condition', 'Samples')
              metadata_new = rbind(controls_df, treatment_df)
              metadata_new = data.frame(metadata_new)
              rownames(metadata_new) = metadata_new$Samples
              mat = assay(dds)
              mat = mat[,rownames(metadata_new)]
              dds_use = DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata_new, design= ~condition)
              keep <- rowSums( counts(dds_use) >= countsfilter ) >= min(length(input$controls), length(input$treatments))
              dds_use <- dds_use[keep,]
              colData(dds_use)[,'condition'] <- relevel(colData(dds_use)[,'condition'], ref = 'Control')
              incProgress(0.5, detail = "Running DESeq2...")
              dds_use <- DESeq(dds_use)      
            }
            incProgress(0.4, detail = "Finalizing Analysis ..")
            shinyalert("DE Analysis is complete", "Let's obtain DE genes by selecting coefficients", type = "success", confirmButtonCol = "#1c82e0")
          }, error = function(e) {
            shinyalert("DE Analysis Failed", "Please check input data", type = "error", confirmButtonCol = "#1c82e0")
          })
        })
      }
      if(input$sample_select == 'Perform DE based on condition column in metadata') {
        req(input$controls_set, input$treatment_set)
        if((!is.null(condition_column)) && (!is.null(input$controls_set)) && (!is.null(input$treatment_set))) {
          withProgress(message = 'Performing DE Analysis...', value = 0, {
            incProgress(0.1, detail = "Checking Data and parameters...")
            tryCatch({
              samples_user = c(input$controls_set, input$treatment_set)
              metadata_filter = metadata %>% dplyr::filter(metadata[[condition_column]] %in% samples_user)
              mat = assay(dds)
              mat = mat[,rownames(metadata_filter)]
              column_name <-  paste("~", condition_column)
              dds_use <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata_filter, design= as.formula(column_name))
              table_condition = data.frame(table(metadata_filter[[condition_column]]))
              colnames(table_condition) = c("condition", "number_of_samples")
              list_number_of_samples = c()
              for(x in 1:length(table_condition$condition))
              {
                val = table_condition$number_of_samples[x]
                list_number_of_samples = c(list_number_of_samples, val)
              }
              keep <- rowSums( counts(dds_use) >= countsfilter ) >= min(list_number_of_samples)
              dds_use <- dds_use[keep,]
              colData(dds_use)[,condition_column] <- relevel(colData(dds_use)[,condition_column], ref = input$controls_set)
              incProgress(0.5, detail = "Running DESeq2...")
              dds_use <- DESeq(dds_use)
              incProgress(0.4, detail = "Finalizing Analysis ..")
              shinyalert("DE Analysis is complete", "Let's obtain DE genes by selecting coefficients", type = "success", confirmButtonCol = "#1c82e0")
            }, error = function(e) {
              shinyalert("DE Analysis Failed", "Please check input data", type = "error", confirmButtonCol = "#1c82e0")
            })
            
          })
        }
      }
      return(dds_use)
    })
    
    observe({
      req(DE_results())
      de_res = DE_results()
      if(!is.null(de_res)) {
        updateSelectizeInput(session, "coefficients", choices = resultsNames(de_res)[[2]], selected = resultsNames(de_res)[[2]])
      }
    })
    
    output$de_table = DT::renderDataTable({
      req(DE_results(), input$coefficients, dataObject())
      dds_original = dataObject()[[1]]
      de_res = DE_results()
      res <- results(de_res, name=input$coefficients)
      res_out = data.frame(res)
      if(input$sample_select == 'Select Samples Manually for comparison') {
        res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter | log2FoldChange <= -1*(input$logFC_filter), padj <= input$fdr_filter) %>% column_to_rownames('gene')
      }
      if(input$sample_select == 'Perform DE based on condition column in metadata') {
        res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter_cond | log2FoldChange <= -1*(input$logFC_filter_cond), padj <= input$fdr_filter_cond) %>% column_to_rownames('gene')
      }
      res_out = arrange(res_out, padj)
      out_list = list(dds_original, de_res, res, res_out, input$logFC_filter, input$fdr_filter, input$coefficients, dataObject()[[4]])
      data(out_list)
      DT::datatable(res_out)
    })
    
    observe({
      if(!is.null(DE_results())) {
        shinyjs::show("downloadde_table")
      }
    })
    
    observe({
      if(!is.null(input$controls) & !is.null(input$treatments)) {
        shinyjs::show("download_condition_table_manual")
      }
    })
    
    output$downloadde_table <- downloadHandler(
      filename = function() {
        paste0("CogentDS_BulkRNA_DE_Results", "-", Sys.Date(), ".csv")
      },
      content = function(file) {
        req(DE_results(), input$coefficients)
        de_res = DE_results()
        res <- results(de_res, name=input$coefficients)
        res_out = data.frame(res)
        if(input$sample_select == 'Select Samples Manually for comparison') {
          res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter | log2FoldChange <= -1*(input$logFC_filter), padj <= input$fdr_filter) %>% column_to_rownames('gene')
        }
        if(input$sample_select == 'Perform DE based on condition column in metadata') {
          res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter_cond | log2FoldChange <= -1*(input$logFC_filter_cond), padj <= input$fdr_filter_cond) %>% column_to_rownames('gene')
        }
        res_out = arrange(res_out, padj)
        write.csv(res_out, file, quote=F)
      }
    )
    
    output$download_condition_table_manual <- downloadHandler(
      filename = function() {
        paste0("CogentDS_BulkRNA_DE_ManualSelection_Table", "-", Sys.Date(), ".csv")
      },
      content = function(file) {
        req(input$controls, input$treatments)
        if(!is.null(input$controls) & !is.null(input$treatments)) {
          control_set_use = rep("Control", length(input$controls))
          treatment_set_use = rep("Treatment", length(input$treatments))
          controls_df = cbind(control_set_use, input$controls)
          colnames(controls_df) = c('condition', 'Samples')
          treatment_df = cbind(treatment_set_use, input$treatments)
          colnames(treatment_df) = c('condition', 'Samples')
          metadata_new = rbind(controls_df, treatment_df)
          metadata_new = data.frame(metadata_new)
          rownames(metadata_new) = metadata_new$Samples
          write.csv(metadata_new, file, quote=F)
        }
      }
    )
    return(data)
  })
}