####### load packages ##########
library(devtools)
library(shiny)
library(shinydashboard)
library(shinyjs)
library(Seurat)
library(SingleCellExperiment)
library(SingleR)
library(decontX)
library(plotly)
library(shinyalert)
library(dplyr)
library(presto)
library(shinycssloaders)
library(DT)
library(celldex)
library(data.table)
library(Matrix)
library(shinyBS)
library(org.Hs.eg.db)
library(org.Mm.eg.db)
library(enrichplot)
library(ggpubr)
library(rlist)
library(clusterProfiler)
library(tidyverse)
library(R.utils)
library(rmarkdown)
library(kableExtra)
library(Routliers)
library(ggrepel)
library(DESeq2)
library(MAST)
library(apeglm)
library(pheatmap)
library(EnhancedVolcano)
library(RColorBrewer)
library(ReactomePA)
library(msigdbr)
library(glmGamPoi)
##############################

# Increase size limit for data
options(shiny.maxRequestSize=10000*1024^2)
#########################################

# Source the R scripts containing UI definitions and server logic for each module of the application
source('R/setup_functions.R')
##############################

# Define UI
ui <- dashboardPage(
  dashboardHeader(title = 'Cogent™ NGS Discovery Software - BulkRNA', titleWidth = 400,
                            tags$li(a(href = 'https://www.takarabio.com/', img(src = "logo.png", title = "Company Home", height = "30px"), style = "padding-top:10px; padding-bottom:10px;"), class = "dropdown"),
                            tags$li(class = "dropdown", actionButton("refreshapp", "Home", icon = icon("home"), style='color: black; background-color: white; border: transparent'))),
    skin = "purple",
  dashboardSidebar(
    disable = F,
    width = 180,
    sidebarMenuOutput("CogentDSsidebarmenu")
  ),

  dashboardBody(
    shinyjs::useShinyjs(),
    tags$head(tags$style(".sidebar-menu li { margin-bottom: 0px; margin-top: 0px;}")),
    tags$head(tags$style(".sidebar-menu li { padding: 0px 0px 0px 0px; display: block}")),
    tags$head(tags$style(HTML('

         .wrapper {height: auto !important; position:relative; overflow-x:hidden; overflow-y:hidden}

         .skin-purple .main-header .logo {
          background-color: white;
          color: black;
          border-bottom-style: solid;
          text-align: left;
          padding: 0px 0px 0px 10px;
          font-style: normal;
          font-family: math;
        }

        .skin-purple .main-header .navbar{
          background-color: white;
          border-bottom: #e4e4e4 solid;
        }

        .skin-purple .main-header .navbar .sidebar-toggle {
          background-color: white;
          color: black;
        }

        /* main sidebar */
          .skin-blue .main-sidebar {
            background-color: #e4e4e4;
        }

        .skin-purple .left-side, .skin-purple .main-sidebar, .skin-purple .wrapper {
          background-color: #e4e4e4;
          border-left-color: #e4e4e4;
        }

        .skin-purple .sidebar-menu>li.active>a, .skin-purple .sidebar-menu>li:hover>a {
          color: #1c82e0;
          text-align: left;
          background-color: #e4e4e4;
          border-left-color: #e4e4e4;
        }

        .skin-purple .main-header .navbar .sidebar-toggle:hover {
          background-color: white
        }

        .skin-purple .main-header .logo:hover {
          background-color: white
        }

        .skin-purple .sidebar a {
          color: darkgreen;
        }

        /* body */
          .content-wrapper, .right-side {
          background-color: white;
        }
    '))),
    tabItems(
      tabItem(tabName = "landingUI", landingUI("landingUIModule")),
      tabItem(tabName = "dataUpload", dataUploadUI("dataUploadModule"), actionButton("gotoqc", "Next: Check Data Quality", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "qc", qcUI("qcModule"), actionButton("gotode", "Next: Perform Differential Expression", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "DE_Analysis", deanalysisUI("DEModule"), actionButton("gotoviz", "Next: Visualization", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "Visualization", visualizationUI("VisualizationModule"), actionButton("gotope", "Next: Pathway Enrichment", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "PathwayEnrichment", enrichmentUI("PathwayErichmentModule"), actionButton("gotoReport", "Go to Analysis Report", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "AnalysisReport", generateReport_UI("generateReportModule"), actionButton("gotodownload", "Go to Download Page", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "DownloadDESeq", downloadDESeq2_UI("DownloadDESeqModule"), actionButton("gotoMain", "Go to Main Page", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0"))

    )
  )
)


# Server Logic
server <- function(input, output, session) {
  
  ns <- session$ns

  ########## Load Server functions

  output$workflow <- renderImage({
  list(src = 'www/workflow.png',
         contentType = 'image/png', height = 600, alt = "This is alternate text")
  }, deleteFile = F)


  output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
  sidebarMenu(id='CogentDSsidebarmenu',
              menuItem("BulkRNA", tabName = "landingUI", icon = icon("dashboard"), selected = "landingUI"))

  })

  observeEvent(input$refreshapp, {
    shinyjs::refresh()
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
          menuItem("BulkRNA", tabName = "landingUI", icon = icon("dashboard")))

    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "landingUI")
  })

  observeEvent(input$confirm, {
    if(!is.null(dev.list())) {
      dev.off()
      message("Found an open Device! setting it off")
    } else {
      message("no open device found, internal test!")
    }
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
          menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload"), selected = "dataUpload"))    
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "dataUpload")
    
  })

  
  observeEvent(input$gotoqc, {
    if(!is.null(dev.list())) {
      dev.off()
      message("Found an open Device! setting it off")
    } else {
      message("no open device found, internal test!")
    }
     output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")))
      })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "qc")
  })

  observeEvent(input$gotode, {
     output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")),
        menuItem("DE Analysis", tabName = "DE_Analysis", icon = icon("chart-line")))
      })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "DE_Analysis")
  })

  observeEvent(input$gotoviz, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")),
        menuItem("DE Analysis", tabName = "DE_Analysis", icon = icon("chart-line")),
        menuItem("DE Visualization", tabName = "Visualization", icon = icon("chart-line")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "Visualization")
  })

  observeEvent(input$gotope, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")),
        menuItem("DE Analysis", tabName = "DE_Analysis", icon = icon("chart-line")),
        menuItem("DE Visualization", tabName = "Visualization", icon = icon("chart-line")),
        menuItem("Pathway Enrichment", tabName = "PathwayEnrichment", icon = icon("chart-line")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "PathwayEnrichment")
  })

  observeEvent(input$gotoReport, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")),
        menuItem("DE Analysis", tabName = "DE_Analysis", icon = icon("chart-line")),
        menuItem("DE Visualization", tabName = "Visualization", icon = icon("chart-line")),
        menuItem("Pathway Enrichment", tabName = "PathwayEnrichment", icon = icon("chart-line")),
        menuItem("Analysis Report", tabName = "AnalysisReport", icon = icon("download")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "AnalysisReport")
  })

  observeEvent(input$gotodownload, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu', 
        menuItem("Upload Data", tabName = "dataUpload", icon = icon("upload")),
        menuItem("QC", tabName = "qc", icon = icon("tachometer-alt")),
        menuItem("DE Analysis", tabName = "DE_Analysis", icon = icon("chart-line")),
        menuItem("DE Visualization", tabName = "Visualization", icon = icon("chart-line")),
        menuItem("Pathway Enrichment", tabName = "PathwayEnrichment", icon = icon("chart-line")),
        menuItem("Analysis Report", tabName = "AnalysisReport", icon = icon("download")),
        menuItem("Download DESeq2", tabName = "DownloadDESeq", icon = icon("download")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "DownloadDESeq")
  })

  observeEvent(input$gotoMain, {
    shinyjs::refresh()
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
          menuItem("BulkRNA", tabName = "landingUI", icon = icon("dashboard")))

    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "landingUI")
  })
  
  dataObject <- dataUploadServer("dataUploadModule")
  qcServer("qcModule", dataObject)
  de_results <- deanalysisServer("DEModule", dataObject)
  visualizationServer("VisualizationModule", de_results)
  enrichmentServer("PathwayErichmentModule", de_results)
  generateReport_Server("generateReportModule", de_results)
  downloadDESeq2_Server("DownloadDESeqModule", de_results)
}

# Run the application
shinyApp(ui = ui, server = server)
