#' UI component for plotting various QC plots
#' 
#' Generates UI elements for a module dedicated to QC plots (Loess, lorenz, gini, MAD plot) within
#' a Shiny application. User can perform filtering based on gini scores, ploidy, samples & barcodes for plotting.
#'  
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return
#' @export
#'
#' @examples
#' GinkgoUI("ginkgoModuleUI")
GinkgoUI <- function(id) {
  ns <- NS(id)
  tagList(
    h2('Visualize QC Plots'),
    
    fluidRow(
      column(4,
             # Input for outlier detection threshold
             numericInput(ns("niqrs"), "Outlier detection threshold", value = 3),
             bsPopover(ns("niqrs"), "Outlier Detection Threshold", 
                       "Adjust the threshold to filter cells based on Gini scores.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Dynamic input for filtering based on sample, ploidy, and barcode
             selectInput(ns("sample"), "Select Sample:", choices = NULL, multiple = TRUE),
             bsPopover(ns("sample"), "Select Sample", 
                       "Filter by samples.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             sliderInput(ns("ploidy"), "Select Ploidy Range:", 
                         min = 0, max = 1, 
                         value = c(0, 1), step = 0.1),
             bsPopover(ns("ploidy"), "Ploidy Range", 
                       "Set the range of ploidy values to filter the data.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             selectizeInput(ns("barcode"), "Select Barcode:", choices = NULL, multiple = TRUE),
             bsPopover(ns("barcode"), "Select Barcode", 
                       "Filter data by specific barcodes.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Select plot type
             selectInput(ns("plot_type"), "Choose a Plot:", selected = NULL,
                         choices = list("Gini Plot" = "Gini_Plot", 
                                        "Loess Plot" = "Loess_Plot", 
                                        "Lorenz plot" = "lorenz_plot",
                                        "MAD Plot (By Samples)" = "MAD_Plot_by_samples",
                                        "MAD Plot" = "MAD_plot")
             ),
             actionButton(ns("plotButton"), "Generate Plot", style='color:#fff; background-color:#1c82e0')
      ),
      
      column(8,
             selectInput(ns("plot_file_type"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
             downloadButton(ns('download_plot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
             shinycssloaders::withSpinner(plotOutput(ns("plot"), width = "600px"))
      
      )
    )
  )
}