#' Server logic for Clustering and Visualization Module
#' 
#' Implements uwot based umap and hdbscan clustering.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param ccnObject Object consisting of GC content, raw data, SegCopy, bounds data (obtained from Ginkgo)
#'
#' @return
#' @export
#'
#' @examples
#' umap_data <- clusteringServer("clusteringModuleUI", data)
clusteringServer <- function(id, ccnObject) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    
    shinyjs::hide("download_plot")
    shinyjs::hide("plot_file_type")
    shinyjs::hide("download_clusters")
    
    data_clustering <- reactiveVal(NULL)
    ccnObject_filtered_cl <- reactiveVal(NULL)
    
    # Filtering based on gini scores
    gini_scores_df_filt <- reactive({
      req(ccnObject()$gini_scores, input$niqrs)
      n_iqrs <- input$niqrs
      
      is_outlier <- function(x, top_only=TRUE) {
        if (top_only) {
          return(x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        } else {
          return(x < quantile(x, 0.25, na.rm = TRUE) - n_iqrs * IQR(x, na.rm = TRUE) | x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
        }
      }
      
      cell_ids <- rownames(ccnObject()$gini_scores)
      gini_scores_df <- as.data.frame(ccnObject()$gini_scores %>% group_by(sample) %>% mutate(outlier = ifelse(is_outlier(gini), TRUE, FALSE)))
      rownames(gini_scores_df) <- cell_ids
      cells_to_remove <- rownames(subset(gini_scores_df, outlier == TRUE))
      list(gini_scores_with_outliers = gini_scores_df, cells_to_remove = cells_to_remove)
    })
    
    # CCN data
    ccn_data <- reactive({
      req(ccnObject(), gini_scores_df_filt()$cells_to_remove)
      ccn_medians <- data.frame(ploidy=colMedians(as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts)))
      cl_names <- gsub("_trimmed$", "", rownames(ccn_medians))
      ccn_medians$sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
      ccn_medians$barcode <- rownames(ccn_medians)
      ccn_medians
    })
    
    # Merge ccn data with gini scores dataframe
    merged_data <- reactive({
      req(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data())
      df <- gini_scores_df_filt()$gini_scores_with_outliers
      df$barcode <- rownames(df)
      # Merging
      merged <- merge(df, ccn_data(), by = "barcode")
      merged <- merged[,!names(merged) %in%  "sample.y"]
      colnames(merged)[which(names(merged) == "sample.x")] <- "sample"
      merged
    })
    
    # Dynamically populate choices for sample, ploidy, and barcode based on merged data
    observe({
      updateSelectInput(session, "sample", choices = unique(merged_data()$sample))
      # Update ploidy slider based on data
      updateSliderInput(session, "ploidy", min = min(merged_data()$ploidy, na.rm = TRUE),
                        max = max(merged_data()$ploidy, na.rm = TRUE),
                        value = c(min(merged_data()$ploidy, na.rm = TRUE), 
                                  max(merged_data()$ploidy, na.rm = TRUE)))
    })
    
    # Update barcodes dynamically based on the selected sample and ploidy filters
    observe({
      filtered_data <- merged_data()
      
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        filtered_data <- filtered_data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter
      filtered_data <- filtered_data %>% filter(ploidy >= input$ploidy[1], 
                                                ploidy <= input$ploidy[2])
      
      # Update barcode choices
      updateSelectizeInput(session, "barcode", choices = unique(filtered_data$barcode), server = TRUE)
    })
    
    # Reactive expression to filter the merged data based on user input for sample, ploidy, and barcode
    filtered_merged_data <- reactive({
      data <- merged_data()
      
      # Apply sample filter
      if (!is.null(input$sample) && length(input$sample) > 0) {
        data <- data %>% filter(sample %in% input$sample)
      }
      
      # Apply ploidy filter
      data <- data %>% filter(ploidy >= input$ploidy[1], 
                              ploidy <= input$ploidy[2])
      
      # Apply barcode filter
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        data <- data %>% filter(barcode %in% input$barcode)
      }
      data
    })
    
    # Reactive expression to filter matrix based on user input
    filtered_matrix <- reactive({
      req(filtered_merged_data(), ccnObject()$ccn_data_files)
      
      # Start with the metadata
      filtered_metadata <- filtered_merged_data()
      ccn_data <- data.frame(as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts))
      ccn_chr_info <- ccnObject()$ccn_data_files[["RNA"]][["Chrom_info"]]
      
      # Split the chromosome info into three new columns
      chr_info <- separate(ccn_chr_info, Chrom_info, into = c("CHR", "START", "END"), sep = "-")
      ccn_data <- data.frame(chr_info, ccn_data)
      
      # Filter by barcode (if provided)
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        filtered_metadata <- filtered_metadata %>% filter(barcode %in% input$barcode)
      }
      
      # Filter for chrY and chrX based on user input
      if (input$include_chromosomes == FALSE) {
        ccn_data <- ccn_data %>% filter(!(CHR %in% c("chrY", "chrX")))
      }
      
      # Get the filtered sample_barcode (corresponds to rownames in CCN data)
      filtered_sample_barcode <- filtered_metadata$barcode
      
      # Subset the matrix columns based on the filtered sample_barcode
      filtered_ccn_data <- ccn_data[, c(1:3, which(colnames(ccn_data) %in% filtered_sample_barcode)), drop = FALSE]
      filtered_ccn_data
    })
    
    plot_data <- reactiveVal(NULL)
    umap_plot <- reactiveVal(NULL)
    
    clustering_data <- eventReactive(input$plot, {
      
      # Alterts 
      
      # Ensure gini_scores_df_filt gets re-evaluated based on the current outlier threshold value
      gini_scores_filtered <- gini_scores_df_filt()
      
      # Alert if no data present based on user selection
      if (all(gini_scores_filtered$gini_scores_with_outliers$outlier==FALSE)) {
        shinyalert("No outliers detected",
                   "The outlier detection threshold did not detect any outliers. Consider adjusting the threshold.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (is.null(ccnObject())) {
        shinyalert('Missing Data', 'Please upload the data.', type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      } 
      
      if (ncol(filtered_matrix()) - 3 <= input$n_neighbors) {
        shinyalert("Error", 
                   "The number of neighbors must be less than the number of available cells. 
             Please reduce the number.", 
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (input$minpts >= ncol(filtered_matrix()) - 3) {
        shinyalert("Error", 
                   "The minimum points must be less than the number of available cells. 
             Please reduce the number.", 
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (input$n_neighbors < 2) {
        shinyalert("Invalid Input", "The number of neighbors (n_neighbors) should be at least 2. Please adjust the value.", 
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (input$minpts < 2) {
        shinyalert("Invalid Input", "The minimum points should be at least 2. Please adjust the value.", 
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Parameter tracking
      updated_ccnObject <- ccnObject()
      updated_ccnObject$user_parameter_tracker <- NULL
      updated_ccnObject$user_parameter_tracker <- c(
        updated_ccnObject$user_parameter_tracker,
        paste("Outlier detection threshold (nIQRs):", input$niqrs),
        paste("Selected Sample(s):", paste(input$sample, collapse = ", ")),
        paste("Selected Ploidy:", paste(input$ploidy, collapse = ", ")),
        paste("Selected Barcode(s):", paste(input$barcode, collapse = ", ")),
        paste("Selected n_neighbors:", input$n_neighbors),
        paste("Selected minpts:", input$minpts)
      )
      
      # Update the ccn object
      ccnObject(updated_ccnObject)
      data_clustering(ccnObject())
      
      # Alert if no data present based on user selection
      if (ncol(filtered_matrix()) == 3) {
        shinyalert("No Data",
                   "No data available for the selected sample, ploidy or barcode combination.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(filtered_matrix(), input$n_neighbors, input$minpts)
      withProgress(message = paste("Generating", "UMAP"), value = 0, {
        incProgress(1 / 10, detail = "Processing UMAP...")
        Sys.sleep(1)
        ccn_data <- filtered_matrix()
        # Apply updated outlier filtering based on the updated niqrs value
        indx <- which(colnames(ccn_data) %in% gini_scores_filtered$cells_to_remove)
        # If no cells to remove, consider all CCN data
        if (length(indx) == 0) {
          subset_data <- ccn_data
        } else {
          subset_data <- ccn_data[, -indx]
        }
        set.seed(1)
        umap_data <- as.matrix(subset_data[, 4:ncol(subset_data)])
      
        # Ensure enough data remains
        if (ncol(umap_data) < input$n_neighbors) {
          shinyalert("Error", 
                     paste0("The number of neighbors ", input$n_neighbors, 
                            " must be smaller than the number of available cells ", ncol(umap_data), 
                            ". Please reduce the number."),
                     type = "error", confirmButtonCol = "#1c82e0")
          return(NULL)
        }
        
        cl_names <- gsub("_trimmed$", "", colnames(umap_data))
        sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
        
        barcode <- colnames(umap_data)
        set.seed(1)
        udata <- uwot::umap(t(umap_data), n_neighbors = input$n_neighbors)
        udata <- data.frame(udata, sample = sample, barcode = barcode)
        colnames(udata) <- c("UMAP1", "UMAP2", "sample", "barcode")
        
        # Perform DBSCAN clustering on UMAP coordinates
        set.seed(1)
        dbscan_res <- dbscan::hdbscan(udata[, c("UMAP1", "UMAP2")], minPts = input$minpts)
        udata$cluster <- as.factor(dbscan_res$cluster)
        
        # Save UMAP and clustering data in Seurat object
        ccn_filtered_barcodes <- udata$barcode
        ccnObject_filtered <- subset(ccnObject()$ccn_data_files, cells = ccn_filtered_barcodes)

        # Save UMAP coordinates
        ccnObject_filtered@reductions$umap <- CreateDimReducObject(
          embeddings = as.matrix(udata[, c("UMAP1", "UMAP2")]),
          key = "UMAP_",
          assay = DefaultAssay(ccnObject_filtered)
        )
        
        # Save clustering, barcode & samples information
        ccnObject_filtered@meta.data$cluster <- udata$cluster
        ccnObject_filtered@meta.data$sample <- udata$sample
        ccnObject_filtered@meta.data$barcode <- udata$barcode
        
        # update ccn object
        ccnObject_filtered_cl(ccnObject_filtered) 
        plot_data(udata)
        return(udata)
      })
    })
    
    # Full UMAP data (no filtering) to be used in lasso module
    full_umap_data <- reactive({
      req(ccnObject())  # Ensure the CCN object is available
      full_ccn_data <- as.matrix(ccnObject()$ccn_data_files@assays[["RNA"]]$counts)
      cl_names <- gsub("_trimmed$", "", colnames(full_ccn_data))
      sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
      barcode <- colnames(full_ccn_data)
      
      # Adjust n_neighbors and minPts for UMAP and Clustering respectively
      n_neighbors_default <- min(15, ncol(full_ccn_data) - 1)
      minPts_default <- min(10, ncol(full_ccn_data) - 1)
      
      # Ensure at least 2 neighbors and minPts are set for UMAP & clustering
      if (n_neighbors_default < 2) {
        message("Insufficient samples for UMAP: Setting n_neighbors to 2.")
        n_neighbors_default <- 2
      }
      if (minPts_default < 2) {
        message("Insufficient samples for DBSCAN: Setting minPts to 2.")
        minPts_default <- 2
      }
      
      set.seed(1)
      umap_full <- uwot::umap(t(full_ccn_data), n_neighbors = n_neighbors_default)
      
      # Create a data frame for the UMAP results
      umap_full_data <- data.frame(umap_full, sample = sample, barcode = barcode)
      colnames(umap_full_data) <- c("UMAP1", "UMAP2", "sample", "barcode")

      # Perform DBSCAN clustering on UMAP coordinates
      set.seed(1)
      dbscan_res <- dbscan::hdbscan(umap_full_data[, c("UMAP1", "UMAP2")], minPts = minPts_default)
      umap_full_data$cluster <- as.factor(dbscan_res$cluster)
      
      # Return the full UMAP data for use in the next module
      umap_full_data
    })
    
    output$UMAP_plot <- renderPlotly({
      req(clustering_data(), input$color_by)
      color_by <- ifelse(input$color_by == "cluster", "cluster", "sample")
      set.seed(1)
      p <- ggplot(clustering_data(), aes(x = UMAP1, y = UMAP2, color = .data[[color_by]], text = paste("barcode", barcode, sep = ":"))) +
        geom_point(size = input$ptSize) + theme_classic()
      umap_plot(p)
      set.seed(1)
      ggplotly(p)
    })
    
    observe({
      if (!is.null(umap_plot())) {
        shinyjs::show("download_plot")
        shinyjs::show("plot_file_type")
        shinyjs::show("download_clusters")
      }
    })
    
    output$download_plot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_", "Clustering", "_", Sys.Date(), input$plot_file_type)
      },
      content = function(file) {
        ggsave(file, plot = umap_plot(), w=4, h=4, units="in", bg="white")
      },
      contentType = 'image'
    )
    
    output$download_clusters <- downloadHandler(
      filename = function() {
        paste0("CogentDS_", "cluster_data", "_", Sys.Date(), ".csv")
      },
      content = function(file) {
        req(umap_plot())
        # Extract only the cell (rownames) and cluster_id
        udata <- clustering_data()
        udata_subset <- data.frame(
          cell = rownames(udata),
          cluster_id = udata$cluster
        )
        
        write.csv(udata_subset, file, row.names = F)
      }
    )
    return(list(full_umap_data = full_umap_data, 
                user_param_tracker_clustering = data_clustering,
                filtered_object = ccnObject_filtered_cl))
    
  })
}
