#' UI logic for Clustering and Visualization Module
#'
#' Generates UI elements for a module dedicated to clustering and visualization within
#' a Shiny application. User can perform filtering based on gini scores, ploidy, samples & barcodes 
#' 
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return
#' @export
#'
#' @examples
#' clusteringUI("clusteringModuleUI")
clusteringUI <- function(id) {
  ns <- NS(id)
  tagList(
    h2("Clustering & Non-linear dimension reduction"), 
    fluidRow(
      column(4,
             
             # Include ChrY/ChrX for clustering or not
             radioButtons(ns("include_chromosomes"), "Include chrX and chrY?",
                          choices = list("TRUE" = TRUE, "FALSE" = FALSE),
                          selected = FALSE),
             bsPopover(ns("include_chromosomes"), "Include chrX and chrY?",
                       "Include chrX and chrY chromosomes in clustering analysis?", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Input for outlier detection threshold
             numericInput(ns("niqrs"), "Outlier detection threshold", value = 3),
             bsPopover(ns("niqrs"), "Outlier Detection Threshold", 
                       "Adjust the threshold to filter cells based on Gini scores.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             
             # Dynamic input for filtering based on sample, ploidy, and barcode
             selectInput(ns("sample"), "Select Sample:", choices = NULL, multiple = TRUE),
             bsPopover(ns("sample"), "Select Sample", 
                       "Filter by samples.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             sliderInput(ns("ploidy"), "Select Ploidy Range:", 
                         min = 0, max = 1, 
                         value = c(0, 1), step = 0.1),
             bsPopover(ns("ploidy"), "Ploidy Range", 
                       "Set the range of ploidy values to filter the data.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             selectizeInput(ns("barcode"), "Select Barcode:", choices = NULL, multiple = TRUE),
             bsPopover(ns("barcode"), "Select Barcode", 
                       "Filter data by specific barcodes.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Parameters for UMAP
             numericInput(ns("n_neighbors"), "Local neighborhood size", value =  15),
             bsPopover(ns("n_neighbors"), "Local neighborhood size", 
                       "Number of neighbors to consider for UMAP embedding.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Parameter for clustering (dbscan)
             numericInput(ns("minpts"), "Minimum cluster size", value =  10),
             bsPopover(ns("minpts"), "Minimum cluster size", 
                       "Number of points required to detect the minimum cluster size for DBSCAN based clustering.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Color code UMAP based on samples or clusters
             selectInput(ns("color_by"), "Color by", choices = c("sample", "cluster"), selected = "sample"),
             bsPopover(ns("color_by"), "Color by", 
                       "Choose whether to color UMAP by samples or clusters.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             # Control point size
             numericInput(ns("ptSize"), "Point size", value = 0.5),
             bsPopover(ns("ptSize"), "Point size", 
                       "Adjust size of points on the UMAP plot.", 
                       placement = "right", trigger = "hover", options = list(container = "body")),
             
             
             # Action Btn for plotting
             actionButton(ns("plot"), "Perform non-linear dimension reduction and clustering", style='color:#fff; background-color:#1c82e0'),
             
      ),
      
      column(8,
             selectInput(ns("plot_file_type"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
             downloadButton(ns('download_plot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
             downloadButton(ns('download_clusters'), 'Download Clusters', style='color:#fff; background-color:#1c82e0'),
             shinycssloaders::withSpinner(plotlyOutput(ns("UMAP_plot"), width = "600px", height = "500px"))
      )
    )
  )
}