# Define the server function
heatmapServer <- function(id, cnv_files) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    
    # Reactive expression to load the ginko data
    ginkgo_data <- reactive({
      req(req(cnv_files$GC(), cnv_files$ginkgo_data()))
      ginkgo_data_m <- reshape2::melt(cnv_files$ginkgo_data())
      ginkgo_data_m$sample <- gsub("_[^_]*$", "", ginkgo_data_m$variable)
      ginkgo_data_normalized <- sweep(cnv_files$ginkgo_data() + 1, 2, colMeans(cnv_files$ginkgo_data() + 1), '/')
      ginkgo_data_normalized <- cbind(cnv_files$GC(), ginkgo_data_normalized)
      nrow(cnv_files$GC()) == nrow(ginkgo_data_normalized)
      ginkgo_data_normalized_m <- reshape2::melt(ginkgo_data_normalized, id.vars="GC_Content")
      ginkgo_data_normalized_m$sample <- gsub("_[^_]*$", "", ginkgo_data_normalized_m$variable)
      gini_scores <- apply(cnv_files$ginkgo_data(), MARGIN=2, FUN=Gini)
      gini_scores_df <- data.frame(sample=gsub("_[^_]*$", "", names(gini_scores)), gini=gini_scores)
      list(gini_scores = gini_scores_df, normalized_data_melted = ginkgo_data_normalized_m, ginko_data_melted = ginkgo_data_m, normalized_data = ginkgo_data_normalized)
    })
    
    # Filtering
    gini_scores_df_filt <- reactive({
      req(ginkgo_data()$gini_scores, input$niqrs)
      n_iqrs <- input$niqrs
      
      is_outlier <- function(x, top_only=TRUE) {
        if (top_only) {
          return(x > quantile(x, 0.75) + n_iqrs * IQR(x))
        } else {
          return(x < quantile(x, 0.25) - n_iqrs * IQR(x) | x > quantile(x, 0.75) + n_iqrs * IQR(x))
        }
      }
      
      cell_ids <- rownames(ginkgo_data()$gini_scores)
      gini_scores_df <- as.data.frame(ginkgo_data()$gini_scores %>% group_by(sample) %>% mutate(outlier = ifelse(is_outlier(gini), TRUE, FALSE)))
      rownames(gini_scores_df) <- cell_ids
      cells_to_remove <- rownames(subset(gini_scores_df, outlier == TRUE))
      outlier_table <- table(subset(gini_scores_df, outlier == TRUE)$sample)
      list(gini_scores_with_outliers = gini_scores_df, cells_to_remove = cells_to_remove)
    })
    
    # CCN data
    ccn_data <- reactive({
      req(cnv_files$ccn(), gini_scores_df_filt()$cells_to_remove)
      ccn_medians <- data.frame(ploidy=colMedians(as.matrix(cnv_files$ccn()[, 4:ncol(cnv_files$ccn())])))
      ccn_medians$sample <- as.factor(gsub("_[ACTG]+$", "", rownames(ccn_medians)))
      ccn_medians$barcode <- rownames(ccn_medians)
      #ccn_medians <- subset(ccn_medians, ! barcode %in% gini_scores_df_filt()$cells_to_remove)
      ccn_medians
    })
    
    # Merge ccn data with gini scores dataframe on the sample
    merged_data <- reactive({
      req(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data())
      
      # Merge based on sample
      merged <- merge(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data(), by = "sample")
      merged
    })
    
    # Dynamically populate choices for sample, ploidy, and barcode based on merged data
    observe({
      updateSelectInput(session, "sample", choices = unique(merged_data()$sample))
      updateSelectInput(session, "ploidy", choices = c("", unique(merged_data()$ploidy)))
      #updateSelectInput(session, "barcode", choices = unique(merged_data()$barcode))
    })
    
    # Reactive expression to filter matrix based on user input
    filtered_mdata_sample_ploidy <- eventReactive(input$plot,{
      req(merged_data())
      # Metadata
      filtered_mdata <- merged_data()
      
      # Apply filters conditionally based on the user's input
      # Filter by sample (if provided)
      if (!is.null(input$sample) && length(input$sample) > 0) {
        filtered_mdata <- filtered_mdata %>% filter(sample %in% input$sample)
      }
      
      # Filter by ploidy (if provided)
      if (!is.null(input$ploidy)) {
        filtered_mdata <- filtered_mdata %>% filter(ploidy >= input$ploidy)
      }
      
      # Filter by barcode (if provided)
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        filtered_mdata <- filtered_mdata %>% filter(barcode %in% input$barcode)
      }
      
      filtered_mdata
    })
    
    # Dynamically populate choices for sample, ploidy, and barcode based on merged data
    observe({
      updateSelectInput(session, "barcode", choices = unique(filtered_mdata_sample_ploidy()$barcode))
    })
    
    
    # Reactive expression to filter matrix based on user input
    filtered_matrix <- eventReactive(input$plot, {
      req(filtered_mdata_sample_ploidy(), cnv_files$ccn())
      # Start with the metadata
      filtered_metadata <- filtered_mdata_sample_ploidy()
      ccn_data <- cnv_files$ccn()
      
      
      # Filter by barcode (if provided)
      if (!is.null(input$barcode) && length(input$barcode) > 0) {
        filtered_metadata <- filtered_metadata %>% filter(barcode %in% input$barcode)
      }
      
      # Get the filtered sample_barcode (corresponds to rownames in ccn data)
      filtered_sample_barcode <- filtered_metadata$barcode
      
      # Subset the matrix columns based on the filtered sample_barcode
      data <- ccn_data[, c(1:3, which(colnames(ccn_data) %in% filtered_sample_barcode)), drop = FALSE]
      data
    })
    
    # Render plots based on the user select plot type from dropdown
    observeEvent(input$plot, {
      output$heatmap <-  renderPlot({
        
        # Alert if no data present based on user selction
        if (ncol(filtered_matrix()) == 3){
          shinyalert("No Data", 
                     "No data available for the selected sample, ploidy or barcode combination.",
                     type = "error")
          return(NULL)
        }
        
        # Alert if no cells to remove present
        cells_filt <- names(filtered_matrix())[names(filtered_matrix()) %in% gini_scores_df_filt()$cells_to_remove]
        if (length(cells_filt)==0){
          shinyalert("No valid selction", 
                     "Please selct valid input.",
                     type = "error")
          return(NULL)
        }
        
        # Heatmap
        message("CNV Heatmap")
        withProgress(message = paste("Creating", "Heatmap"), value = 0, {
          Sys.sleep(0.25)
          req(filtered_matrix(), gini_scores_df_filt()$cells_to_remove)
          print(filtered_matrix())
          print(dim(filtered_matrix()))
          print(gini_scores_df_filt()$cells_to_remove)
          incProgress(1/10, detail = "Processing Heatmap...")
          Sys.sleep(20)
          cnv_heatmap(filtered_matrix(), horizontal_row_titles = F, show_y_chr = T, cells_to_remove=gini_scores_df_filt()$cells_to_remove)
        })
      })
    })
  })
}