####### load packages ##########
library(dplyr)
library(gglorenz)
library(ggplot2)
library(ineq)
library(MatrixGenerics)
library(magrittr)
library(tidyr)
library(shiny)
library(shinydashboard)
library(shinyalert)
library(InteractiveComplexHeatmap)
library(uwot)
library(plotly)
library(DNAcopy)
library(Seurat)
library(SeuratObject)
library(shinyBS)
##############################

options(Seurat.object.assay.version = "v5")

# Increase size limit for data
options(shiny.maxRequestSize=10000*1024^2)

# Source the R scripts containing UI definitions and server logic for each module of the application
source('R/setup_functions.R')
##############################


# Define UI
ui <- dashboardPage(
  skin = "purple",
  dashboardHeader(title = 'Cogent™ NGS Discovery Software - scDNA', titleWidth = 400,
                  tags$li(a(href = 'https://www.takarabio.com/', img(src = "logo.png", title = "Company Home", height = "30px"), style = "padding-top:10px; padding-bottom:10px;"), class = "dropdown"),
                  tags$li(class = "dropdown", actionButton("refreshapp", "Home", icon = icon("home"), style='color: black; background-color: white; border: transparent'))),
  dashboardSidebar(
    width = 150,
    sidebarMenuOutput("CogentDSsidebarmenu")
  ),
  
  dashboardBody(
    shinyjs::useShinyjs(),
    tags$head(tags$style(".sidebar-menu li { margin-bottom: 0px; margin-top: 0px;}")),
    tags$head(tags$style(".sidebar-menu li { padding: 0px 0px 0px 0px; display: block}")),
    tags$head(tags$style(HTML('

        .wrapper {height: auto !important; position:relative; overflow-x:hidden; overflow-y:hidden}

         .skin-purple .main-header .logo {
          background-color: white;
          color: black;
          border-bottom-style: solid;
          text-align: left;
          padding: 0px 0px 0px 10px;
          font-style: normal;
          font-family: math;
        }

        .skin-purple .main-header .navbar{
          background-color: white;
          border-bottom: #e4e4e4 solid;
        }

        .skin-purple .main-header .navbar .sidebar-toggle {
          background-color: white;
          color: black;
        }

        /* main sidebar */
          .skin-blue .main-sidebar {
            background-color: #e4e4e4;
        }

        .skin-purple .left-side, .skin-purple .main-sidebar, .skin-purple .wrapper {
          background-color: #e4e4e4;
          border-left-color: #e4e4e4;
        }

        .skin-purple .sidebar-menu>li.active>a, .skin-purple .sidebar-menu>li:hover>a {
          color: #1c82e0;
          text-align: left;
          background-color: #e4e4e4;
          border-left-color: #e4e4e4;
        }

        .skin-purple .main-header .navbar .sidebar-toggle:hover {
          background-color: white
        }

        .skin-purple .main-header .logo:hover {
          background-color: white
        }

        .skin-purple .sidebar a {
          color: darkgreen;
        }

        /* body */
          .content-wrapper, .right-side {
          background-color: white;
        }
        
    '))),
    tabItems(
      tabItem(tabName = "landingUI", landingUI("landingUIModule")),
      tabItem(tabName = "DataUpload", DataUploadUI("DataUploadModuleUI"), actionButton("gotoqc", "Next: Visualize QC plots", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "QC_viz", GinkgoUI("ginkgoModuleUI"), actionButton("gotoHeatmap", "Next: Visualize CCN Heatmap", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "cnv_heatmap_viz",
              h2('Visualize Heatmap'),
              fluidRow(
                column(4,
                       # Numeric input for outlier detection based on gini score
                       numericInput("niqrs", "Outlier detection threshold", value = 3),
                       bsPopover(("niqrs"), "Outlier Detection Threshold", 
                                 "Adjust the threshold to filter cells based on Gini scores.", 
                                 placement = "right", trigger = "hover", options = list(container = "body")),
                       
                       # Dynamic input for filtering based on sample, ploidy, and barcode
                       selectInput("sample", "Select Sample:", choices = NULL, multiple = TRUE),
                       bsPopover("sample", "Select Sample", 
                                 "Filter by samples.", 
                                 placement = "right", trigger = "hover", options = list(container = "body")),
                       
                       sliderInput("ploidy", "Select Ploidy Range:", 
                                   min = 0, max = 1, 
                                   value = c(0, 1), step = 0.1),
                       bsPopover("ploidy", "Ploidy Range", 
                                 "Set the range of ploidy values to filter the data.", 
                                 placement = "right", trigger = "hover", options = list(container = "body")),
                       
                       selectizeInput("barcode", "Select Barcode:", choices = NULL, multiple = TRUE),
                       bsPopover("barcode", "Select Barcode", 
                                 "Filter data by specific barcodes.", 
                                 placement = "right", trigger = "hover", options = list(container = "body")),
                       
                       # Button to plot
                       actionButton("heatmap_plot", "Generate plot", style='color:#fff; background-color:#1c82e0'),
                       selectInput("plot_file_type", "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                       downloadButton('download_plot', 'Save plot', style='color:#fff; background-color:#1c82e0'),
                )
              ),
              fluidRow(
                #plotOutput("heatmap")
                InteractiveComplexHeatmapOutput(action = "click",
                                                width1 = 1200,
                                                height1 = 450, compact = T)             
              ), actionButton("gotocn", "Next: Visualize CN profile plot", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")
      ),
      tabItem(tabName = "CNProfile", ccnplotUI("ccnplotModuleUI"), actionButton("gotoclust", "Next: Perform Clustering", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "Clust", clusteringUI("clusteringModuleUI"), actionButton("gotolasso", "Next: Perform Custom Lasso Selection", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "CustomSel", lassoUI("lassoUIModule"), actionButton("gotoreport", "Generate HTML Report", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "generateReport", generateReport_UI("generateReport_Module"), actionButton("gotodownload", "Next: Download Processed Data", class = "btn-primary", style = "position: absolute; bottom: 0; right: 0; color:#fff; background-color:#1c82e0")),
      tabItem(tabName = "download_object", downloadObject_UI("downloadObject_UIModule"))
      
    )
  )
)

# Server Logic
server <- function(input, output, session) {
  
  #ns <- session$ns
  
  output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
    sidebarMenu(id='CogentDSsidebarmenu',
                menuItem("scDNA", tabName = "landingUI", icon = icon("dashboard"), selected = "landingUI"))
    
  })
  
  output$workflow <- renderImage({
    list(src = 'www/workflow.png',
         contentType = 'image/png',
         alt = "This is alternate text", height = 600)
  }, deleteFile = F)
  
  observeEvent(input$refreshapp, {
    shinyjs::refresh()
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("scDNA", tabName = "landingUI", icon = icon("dashboard")))
      
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "landingUI")
  })
  
  observeEvent(input$confirm, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload"), selected = "DataUpload"))    
    })
    
  })
  
  observeEvent(input$gotoqc, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "QC_viz")
  })
  
  observeEvent(input$gotoHeatmap, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "cnv_heatmap_viz")
  })
  
  
  observeEvent(input$gotocn, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")),
                  menuItem("CN Profile", tabName = "CNProfile", icon = icon("chart-bar")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "CNProfile")
  })
  
  
  observeEvent(input$gotoclust, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")),
                  menuItem("CN Profile", tabName = "CNProfile", icon = icon("chart-bar")),
                  menuItem("Clustering", tabName = "Clust", icon = icon("project-diagram")))
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "Clust")
  })
  
  
  observeEvent(input$gotolasso, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")),
                  menuItem("CN Profile", tabName = "CNProfile", icon = icon("chart-bar")),
                  menuItem("Clustering", tabName = "Clust", icon = icon("project-diagram")),
                  menuItem("Custom Selection", tabName = "CustomSel", icon = icon("chart-line"))
      )
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "CustomSel")
  })
  
  observeEvent(input$gotoreport, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")),
                  menuItem("CN Profile", tabName = "CNProfile", icon = icon("chart-bar")),
                  menuItem("Clustering", tabName = "Clust", icon = icon("project-diagram")),
                  menuItem("Custom Selection", tabName = "CustomSel", icon = icon("chart-line")),
                  menuItem("Analysis Report", tabName = "generateReport", icon = icon("download"))
                  
      )
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "generateReport")
  })
  
  
  observeEvent(input$gotodownload, {
    output$CogentDSsidebarmenu <- shinydashboard::renderMenu({
      sidebarMenu(id='CogentDSsidebarmenu',
                  menuItem("Upload Data", tabName = "DataUpload", icon = icon("upload")),
                  menuItem("QC", tabName = "QC_viz", icon = icon("tachometer-alt")),
                  menuItem("Heatmap", tabName = "cnv_heatmap_viz", icon = icon("th")),
                  menuItem("CN Profile", tabName = "CNProfile", icon = icon("chart-bar")),
                  menuItem("Clustering", tabName = "Clust", icon = icon("project-diagram")),
                  menuItem("Custom Selection", tabName = "CustomSel", icon = icon("chart-line")),
                  menuItem("Analysis Report", tabName = "generateReport", icon = icon("download")),
                  menuItem("Download Object", tabName = "download_object", icon = icon("download"))
                  
      )
    })
    updateTabItems(session, "CogentDSsidebarmenu", selected = "download_object")
  })
  
  # server functions
  data <- DataUploadServer("DataUploadModuleUI")
  qcobject <- GinkgoServer("ginkgoModuleUI", data)
  
  shinyjs::hide("download_plot")
  shinyjs::hide("plot_file_type")
  
  heatmap_data <- reactiveVal(NULL)
  heatmap_plot <- reactiveVal(NULL)
  
  # Filtering based on gini scores
  gini_scores_df_filt <- reactive({
    req(data()$gini_scores, input$niqrs)
    n_iqrs <- input$niqrs
    
    is_outlier <- function(x, top_only=TRUE) {
      if (top_only) {
        return(x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
      } else {
        return(x < quantile(x, 0.25, na.rm = TRUE) - n_iqrs * IQR(x, na.rm = TRUE) | x > quantile(x, 0.75, na.rm = TRUE) + n_iqrs * IQR(x, na.rm = TRUE))
      }
    }
    
    cell_ids <- rownames(data()$gini_scores)
    gini_scores_df <- as.data.frame(data()$gini_scores %>% group_by(sample) %>% mutate(outlier = ifelse(is_outlier(gini), TRUE, FALSE)))
    rownames(gini_scores_df) <- cell_ids
    cells_to_remove <- rownames(subset(gini_scores_df, outlier == TRUE))
    outlier_table <- table(subset(gini_scores_df, outlier == TRUE)$sample)
    list(gini_scores_with_outliers = gini_scores_df, cells_to_remove = cells_to_remove)
  })
  
  # CCN data
  ccn_data <- reactive({
    req(data(), gini_scores_df_filt()$cells_to_remove)
    ccn_medians <- data.frame(ploidy=colMedians(as.matrix(data()$ccn_data_files@assays[["RNA"]]$counts)))
    cl_names <- gsub("_trimmed$", "", rownames(ccn_medians))
    ccn_medians$sample <- as.factor(gsub("_[ACTG]+$", "", cl_names))
    ccn_medians$barcode <- rownames(ccn_medians)
    ccn_medians
  })
  
  
  # Merge ccn data with gini scores dataframe
  merged_data <- reactive({
    req(gini_scores_df_filt()$gini_scores_with_outliers, ccn_data())
    df <- gini_scores_df_filt()$gini_scores_with_outliers
    df$barcode <- rownames(df)
    # Merging
    merged <- merge(df, ccn_data(), by = "barcode")
    merged <- merged[,!names(merged) %in%  "sample.y"]
    colnames(merged)[which(names(merged) == "sample.x")] <- "sample"
    merged
  })
  
  # Dynamically populate choices for sample, ploidy, and barcode based on merged data
  observe({
    updateSelectInput(session, "sample", choices = unique(merged_data()$sample))
    # Update ploidy slider range based on data
    updateSliderInput(session, "ploidy", min = min(merged_data()$ploidy, na.rm = TRUE),
                      max = max(merged_data()$ploidy, na.rm = TRUE),
                      value = c(min(merged_data()$ploidy, na.rm = TRUE), 
                                max(merged_data()$ploidy, na.rm = TRUE)))
  })
  
  # Update barcodes dynamically based on the selected sample and ploidy filters
  observe({
    filtered_data <- merged_data()
    
    # Apply sample filter
    if (!is.null(input$sample) && length(input$sample) > 0) {
      filtered_data <- filtered_data %>% filter(sample %in% input$sample)
    }
    
    # Apply ploidy filter based on slider range
    filtered_data <- filtered_data %>% filter(ploidy >= input$ploidy[1], 
                                              ploidy <= input$ploidy[2])
    
    
    # Update barcode choices
    updateSelectizeInput(session, "barcode", choices = unique(filtered_data$barcode), server = TRUE)
  })
  
  
  # Reactive expression to filter the merged data based on user input for sample, ploidy, and barcode
  filtered_merged_data <- reactive({
    data <- merged_data()
    # Apply sample filter
    if (!is.null(input$sample) && length(input$sample) > 0) {
      data <- data %>% filter(sample %in% input$sample)
    }
    
    # Apply ploidy filter
    data <- data %>% filter(ploidy >= input$ploidy[1], 
                            ploidy <= input$ploidy[2])
    
    data
  })
  
  # Reactive expression to filter matrix based on user input
  filtered_matrix <- reactive({
    req(filtered_merged_data(), data()$ccn_data_files)
    # Start with the metadata
    filtered_metadata <- filtered_merged_data()
    ccn_data <- data.frame(as.matrix(data()$ccn_data_files@assays[["RNA"]]$counts))
    ccn_chr_info <- data()$ccn_data_files[["RNA"]][["Chrom_info"]] 
    
    # Split the chromosome info into three new columns
    chr_info <- separate(ccn_chr_info, Chrom_info, into = c("CHR", "START", "END"), sep = "-")    
    ccn_data <- data.frame(chr_info, ccn_data)
    
    # Filter by barcode (if provided)
    if (!is.null(input$barcode) && length(input$barcode) > 0) {
      filtered_metadata <- filtered_metadata %>% filter(barcode %in% input$barcode)
    }
    
    # Get the filtered sample_barcode (corresponds to rownames in ccn data)
    filtered_sample_barcode <- filtered_metadata$barcode
    
    # Subset the matrix columns based on the filtered sample_barcode
    data <- ccn_data[, c(1:3, which(colnames(ccn_data) %in% filtered_sample_barcode)), drop = FALSE]
    data
  })
  
  
  # Render plots based on the user select plot type from dropdown
  observeEvent(input$heatmap_plot, {
    if (is.null(data())) {
      shinyalert('Missing Data', 'Please upload the data.', type = "error", confirmButtonCol = "#1c82e0")
      return(NULL)
    }
    
    # Ensure gini_scores_df_filt gets re-evaluated based on the current outlier threshold value
    gini_scores_filtered <- gini_scores_df_filt()
    
    # Alert if no data present based on user selection
    if (all(gini_scores_filtered$gini_scores_with_outliers$outlier==FALSE)) {
      shinyalert("No outliers detected",
                 "The outlier detection threshold did not detect any outliers. Consider adjusting the threshold.",
                 type = "error", confirmButtonCol = "#1c82e0")
      return(NULL)
    }
    
    req(data())
    
    # Paramter tracking
    updated_ccnObject <- data()
    updated_ccnObject$user_parameter_tracker <- NULL
    updated_ccnObject$user_parameter_tracker <- c(
      updated_ccnObject$user_parameter_tracker,
      paste("Outlier detection threshold (nIQRs):", input$niqrs),
      paste("Selected Sample(s):", paste(input$sample, collapse = ", ")),
      paste("Selected Ploidy:", paste(input$ploidy, collapse = ", ")),
      paste("Selected Barcode(s):", paste(input$barcode, collapse = ", "))
    )
    
    # Update data
    heatmap_data(updated_ccnObject)
    
    # Compute cells to filter
    cells_filt <- intersect(names(filtered_matrix()), gini_scores_df_filt()$cells_to_remove)
    
    # Check if there are no cells flagged for removal
    if (length(cells_filt) == 0) {
      # No cells flagged for removal
      message("No cells flagged for removal. Including all cells in the heatmap.")
      cells_to_remove <- c()  # Empty vector for no removal
      
      # Proceed with heatmap generation
      if (ncol(filtered_matrix()) == 3) {
        shinyalert("No Data",
                   "No data available for the selected sample, ploidy, or barcode combination.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Check if there is sufficient data
      if (ncol(filtered_matrix()) - 3 < 2) {  # Less than 2 barcodes
        shinyalert("Insufficient data: This contains only one barcode. Ensure there are enough barcodes.", 
                   type = "warning", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      message("CNV Heatmap for all cells")
      withProgress(message = paste("Creating", "Heatmap"), value = 0, {
        Sys.sleep(0.25)
        req(filtered_matrix())
        incProgress(1/10, detail = "Processing Heatmap...")
        
        # Generate heatmap for all cells
        hm <- cnv_heatmap(filtered_matrix(), 
                          horizontal_row_titles = F, 
                          show_y_chr = T, 
                          cells_to_remove = cells_to_remove)  # Pass empty vector
        
        heatmap_plot(hm)
        makeInteractiveComplexHeatmap(input, output, session, hm)
        return(heatmap_data)
      })
    } else {
      # Cells are flagged for removal
      message("Generating heatmap with cells flagged for removal.")
      cells_to_remove <- cells_filt 
      
      # Proceed with heatmap generation
      if (ncol(filtered_matrix()) == 3) {
        shinyalert("No Data",
                   "No data available for the selected sample, ploidy, or barcode combination.",
                   type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Check if there are at least 2 barcodes after removing flagged cells
      filtered_matrix_after_removal <- filtered_matrix()[, !colnames(filtered_matrix()) %in% cells_to_remove, drop = FALSE]
      if (ncol(filtered_matrix_after_removal) - 3 < 2) {  # Less than 2 barcodes
        shinyalert("Insufficient data: This contains only one barcode after removing flagged cells. Ensure there are enough barcodes.", 
                   type = "warning", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      message("CNV Heatmap with cell filtering")
      withProgress(message = paste("Creating", "Heatmap"), value = 0, {
        Sys.sleep(0.25)
        req(filtered_matrix(), cells_to_remove)
        incProgress(1/10, detail = "Processing Heatmap...")
        
        # Generate heatmap excluding flagged cells
        hm <- cnv_heatmap(filtered_matrix(), 
                          horizontal_row_titles = F, 
                          show_y_chr = T, 
                          cells_to_remove = cells_to_remove)  # Pass filtered cells
        
        heatmap_plot(hm)
        makeInteractiveComplexHeatmap(input, output, session, hm)
        return(heatmap_data)
      })
    }
  })
  
  observe({
    if (!is.null(heatmap_plot())) {
      shinyjs::show("download_plot")
      shinyjs::show("plot_file_type")
    }
  })
  
  output$download_plot <- downloadHandler(
    filename = function() {
      paste0("CogentDS_", "CNV_traces_heatmap_", Sys.Date(), input$plot_file_type)
    },
    content = function(file) {
      req(heatmap_plot())  # Ensure heatmap_plot() is available
      
      withProgress(message = "Preparing plot for download...", value = 0, {
        
        # Open the appropriate graphics device based on the selected file type
        if (input$plot_file_type == ".png") {
          incProgress(0.2, detail = "Opening PNG device...")
          png(file, width = 14, height = 7, units = "in", res = 300, bg = "white")
          
          incProgress(0.4, detail = "Drawing heatmap plot...")
          draw(heatmap_plot())
          
          incProgress(0.2, detail = "Closing PNG device...")
          dev.off()
          
        } else if (input$plot_file_type == ".pdf") {
          incProgress(0.2, detail = "Opening PDF device...")
          pdf(file, width = 14, height = 7, bg = "white")
          
          incProgress(0.4, detail = "Drawing heatmap plot...")
          draw(heatmap_plot())
          
          incProgress(0.2, detail = "Closing PDF device...")
          dev.off()
          
        } else if (input$plot_file_type == ".jpeg") {
          incProgress(0.2, detail = "Opening JPEG device...")
          jpeg(file, width = 14, height = 7, units = "in", res = 300, bg = "white")
          
          incProgress(0.4, detail = "Drawing heatmap plot...")
          draw(heatmap_plot())
          
          incProgress(0.2, detail = "Closing JPEG device...")
          dev.off()
          
        } else if (input$plot_file_type == ".svg") {
          incProgress(0.2, detail = "Opening SVG device...")
          svg(file, width = 14, height = 7)
          
          incProgress(0.4, detail = "Drawing heatmap plot...")
          draw(heatmap_plot())
          
          incProgress(0.2, detail = "Closing SVG device...")
          dev.off()
        }
        
        # Finalize progress
        incProgress(0.2, detail = "Finalizing download...")
      })
    },
    contentType = "image"
  )
  
  ccnplotServer("ccnplotModuleUI", data)
  umap_data <- clusteringServer("clusteringModuleUI", data)
  lassoServer("lassoUIModule", umap_data$full_umap_data, data)
  generateReport_Server("generateReport_Module", qcobject, heatmap_data, umap_data$user_param_tracker_clustering)
  downloadObject_Server("downloadObject_UIModule", umap_data$filtered_object)
  
}

# Run the application
shinyApp(ui = ui, server = server)
