#' Dimension Reduction Visulaization Server
#' 
#' Handles server-side logic for the additional plots module. This allows users 
#' to visualize Dimension reduction plots 
#'
#' @param id A unique identifier for the server module. This is necessary for 
#'           distinguishing this module's input from others in the same Shiny 
#'           application.
#'           
#' @param seuratObj A reactive expression containing the processed Seurat.
#'
#' @return
#' @export
#' 
#' @examples
#' DiscoverydimServer("DiscoverydimredModule", seuratObject_processed)
#'
DiscoverydimServer <- function(id, seuratObject_processed) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    shinyjs::hide("downloadPlot_umap")
    shinyjs::hide("downloadPlot_pca")
    shinyjs::hide("umapPlt")
    shinyjs::hide("pcaPlt")
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(seuratObject_processed())
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObject_processed()@meta.data)
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObject_processed()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    
    observe({ 
      req(seuratObject_processed())
      updateSelectInput(session, "plotDimPlot_umap_options_group", choices = metadataColumns())
    })
    
    observe({ 
      req(seuratObject_processed())
      updateSelectInput(session, "plotDimPlot_pca_options_group", choices = metadataColumns())
    })
    
    observeEvent(input$plotDimPlot_umap, {
      req(seuratObject_processed(), input$plotDimPlot_umap_options_group)
      tryCatch({
        umapplt = reactive({
          DimPlot(seuratObject_processed(), pt.size = 2, group.by = input$plotDimPlot_umap_options_group, label = TRUE, label.size = 5) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(fill = "white"))
        })
        output$interactivePlot_umap <- renderPlot({
          umapplt()
        })
        shinyalert("UMAP plot generated", "Try other visualizations", type = "success", confirmButtonCol = "#1c82e0")
      }, error = function(e) {
        shinyalert("UMAP plot Failed", "Please check input data and options", type = "error", confirmButtonCol = "#1c82e0")
      }) 
      
      if(!is.null(umapplt()))
      {
        shinyjs::show("downloadPlot_umap")
        shinyjs::show("umapPlt")
      }
      
      output$downloadPlot_umap <- downloadHandler(
        filename = function() {
          paste0("CogentDS_DiscoveryMode_UmapPlot", "-", Sys.Date(), input$umapPlt)
        },
        content = function(file) {
          umapplt()
          ggsave(file, plot = umapplt(), width = 10, height = 10)
        },
        contentType = 'image'
      ) 
    })
    
    observeEvent(input$plotDimPlot_pca, {
      req(seuratObject_processed(), input$plotDimPlot_pca_options_group)
      tryCatch({
        pcaplt = reactive({
          DimPlot(seuratObject_processed(), pt.size = 2, reduction = "pca", group.by = input$plotDimPlot_pca_options_group, label = TRUE, label.size = 5) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(fill = "white"))
        })
        output$interactivePlot_pca <- renderPlot({
          pcaplt()
        })
        shinyalert("PCA plot generated", "Try other visualizations", type = "success", confirmButtonCol = "#1c82e0")
      }, error = function(e) {
        shinyalert("PCA plot Failed", "Please check input data and options", type = "error", confirmButtonCol = "#1c82e0")
      })
      
      if(!is.null(pcaplt()))
      {
        shinyjs::show("downloadPlot_pca")
        shinyjs::show("pcaPlt")
      }
      
      output$downloadPlot_pca <- downloadHandler(
        filename = function() {
          paste0("CogentDS_DiscoveryMode_PCAPlot", "-", Sys.Date(), input$pcaPlt)
        },
        content = function(file) {
          pcaplt()
          ggsave(file, plot = pcaplt(), width = 10, height = 10)
        },
        contentType = 'image'
      ) 
    })
  })
}