#' Server logic for Cell Type Annotation
#'
#' Handles server-side logic for the cell type annotation module. This includes
#' uploading a reference dataset, performing cell type annotation using SingleR,
#' displaying the annotation results in a heatmap, and providing the option to
#' download the annotation results as a CSV file.
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'           
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#' @seealso \code{\link[SingleR]{SingleR}} for more information on cell type annotation.
#' @examples
#' cellTypeAnnotationServer("cellTypeAnnotationModule", seuratObject)
cellTypeAnnotationServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    shinyjs::hide("downloadPlot_CA")
    shinyjs::hide("downloadAnnotationResults")
    shinyjs::hide("labelplot")
    shinyjs::hide("ptSize")
    shinyjs::hide("annoplot")
    
    cellTypeAnnotation <- reactiveValues(annotations = NULL)
    referenceData <- reactiveVal(NULL)
    
    # Define available references by species
    human_references <- c("HumanCellAtlas", "BluePrintEncode", "ImmuneDatabase", "Hematopoietic", "ImmuneMonaco")
    mouse_references <- c("MouseRNAseqData", "ImmGenData")
    
    # Update reference choices based on selected species
    observe({
      available_references <- if (input$species == "Human") human_references else mouse_references
      updateSelectInput(session, "preExistingData", choices = available_references)
    })
    
    # Generate UI elements based on selected reference type and species selection
    observeEvent(input$referenceType, {
      if (input$referenceType == "Upload reference") {
        output$referenceUI <- renderUI({
          fluidRow(
            column(5,
                   fileInput(ns("referenceFile"), "Upload reference RDS file", accept = ".rds"),
                   selectInput(ns("refLabels"), "Reference cell labels column", choices = NULL)
            )
          )
        })
      } else {
        output$referenceUI <- renderUI({
          fluidRow(
            column(5,
                   popify(selectInput(ns("preExistingData"), "Select pre-existing dataset", 
                                      choices = if (input$species == "Human") human_references else mouse_references, width = "500px"), 
                          "Pre-existing datasets", 
                          "Users can select pre-existing datasets from the celldex R package. For more information, refer to the celldex bioconductor page.", 
                          placement = "right", trigger = "hover", options = list(container = "body")),
                   selectInput(ns("refLabels"), "Reference cell labels column", choices = NULL, width = "200px")
            )
          )
        })
      }
    })
    
    # Load reference dataset or pre-existing dataset
    observeEvent(input$referenceFile, {
      req(input$referenceFile)
      referenceData(readRDS(input$referenceFile$datapath))
    })
    
    observeEvent(input$referenceFile, {
      req(input$referenceFile)
      tryCatch({
        reference <- readRDS(input$referenceFile$datapath)
        if (inherits(reference, "Seurat")) {
          reference <- as.SingleCellExperiment(reference, assay = "RNA")
        }
        referenceData(reference)
      }, error = function(e) {
        shinyalert("Error", "Invalid reference file. Please upload a valid RDS file.", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    
    observeEvent(input$preExistingData, {
      withProgress(message = 'Loading reference data...', {
        ref_data <- switch(input$preExistingData,
                           "HumanCellAtlas" = HumanPrimaryCellAtlasData(),
                           "BluePrintEncode" = BlueprintEncodeData(),
                           "ImmuneDatabase" = DatabaseImmuneCellExpressionData(),
                           "Hematopoietic" = NovershternHematopoieticData(),
                           "ImmuneMonaco" = MonacoImmuneData(),
                           "MouseRNAseqData" = MouseRNAseqData(),
                           "ImmGenData" = ImmGenData())
        
        # Rename column names in dropdown
        col_names <- c("label.main" = "Main cell types", 
                       "label.fine" = "Cell subtypes",  
                       "label.ont" = "Cell ontology")
        
        ref_data_renamed <- ref_data
        colnames(colData(ref_data_renamed)) <- col_names[colnames(colData(ref_data))]
        
        referenceData(ref_data_renamed)
      })
    })
    
    # Reactive value to hold metadata columns (factor and character)
    metadataColumns <- reactive({
      req(referenceData())
      
      # Get the names of metadata columns
      all_metadata <- colnames(colData(referenceData()))
      
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(colData(referenceData())[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Return the filtered columns
      relevant_columns
    })
    
    
    # Update dropdown choices based on loaded data
    observe({
      req(referenceData())
      withProgress(message = 'Loading labels...', {
        updateSelectInput(session, "refLabels", choices = metadataColumns())
      })
    })
    
    # Perform cell type annotation on button click
    shiny::observeEvent(input$performCellAnnotations, {
      
      if (input$performCellAnnotations > 0) {
        if (is.null(referenceData()) || is.null(input$refLabels)) {
          shinyalert("Error", "Please upload reference data for cell type annotation.", type = "error", confirmButtonCol = "#1c82e0")
          return()
        }
      }
      
      # Check if "Upload reference" is selected but no file has been uploaded
      if (input$referenceType == "Upload reference" && is.null(input$referenceFile)) {
        shinyalert("Error", "Please upload a reference file before performing cell type annotation.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Checks for selected species again query/ref dataset
      ref_gene_sample <- rownames(referenceData())
      query_gene_sample <- rownames(seuratObj())
      
      # Count occurrences of human and mouse patterns
      ref_human_count <- sum(grepl("^ENSG|^[A-Z]+$", ref_gene_sample))
      ref_mouse_count <- sum(grepl("^ENSMUSG|^[A-Z][a-z]", ref_gene_sample))
      query_human_count <- sum(grepl("^ENSG|^[A-Z]+$", query_gene_sample))
      query_mouse_count <- sum(grepl("^ENSMUSG|^[A-Z][a-z]", query_gene_sample))
      
      # Check against selected species
      if (input$species == "Human" && (ref_mouse_count > ref_human_count || query_mouse_count > query_human_count)) {
        shinyalert("Warning", "The data appears to contain mouse genes while 'Human' is selected.", type = "warning", confirmButtonCol = "#1c82e0")
        return(NULL)
      } else if (input$species == "Mouse" && (ref_human_count > ref_mouse_count || query_human_count > query_mouse_count)) {
        shinyalert("Warning", "The data appears to contain human genes while 'Mouse' is selected.", type = "warning", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(referenceData(), input$refLabels, input$annotation_type)
      
      shinyjs::disable("performCellAnnotations")
      
      # Start timing
      timeStart <- Sys.time()
      
      tryCatch({
        withProgress(message = 'Running SingleR...', {
          incProgress(0.1, detail = 'Loading reference dataset...')
          
          # Simulate the annotation processing time
          Sys.sleep(2) 
          
          incProgress(0.2, detail = 'Converting Seurat to SCE...')
          updatedSeuratObject = seuratObj()
          parameter_list = updatedSeuratObject@assays$user_parameter_tracker
          if(!is.null(input$referenceFile)) {
            reference = paste0('Reference: ', input$referenceFile)
          }
          if(!is.null(input$preExistingData)) {
            reference = paste0('Reference: ', input$preExistingData)
          }
          parameter_list = c(parameter_list, reference)
          updatedSeuratObject@assays$user_parameter_tracker = parameter_list
          query <- as.SingleCellExperiment(updatedSeuratObject, assay = "RNA")
          
          incProgress(0.3, detail = 'Running SingleR...')
          if(input$annotation_type == "Clusters (Fast)") {
            annotations <- SingleR::SingleR(test = query, ref = referenceData(), labels = referenceData()[[input$refLabels]], clusters = updatedSeuratObject$seurat_clusters)

            for(f in 1:nrow(updatedSeuratObject@meta.data)) {
              annotations_temp = annotations[updatedSeuratObject@meta.data$seurat_clusters[f],]
              updatedSeuratObject@meta.data$SingleR_Celltypes[f] = annotations_temp$labels
              updatedSeuratObject@meta.data$SingleR_pruned_Celltypes[f] = annotations_temp$pruned.labels
            }
          }
          if(input$annotation_type == "Cells (Slow)") {
            annotations <- SingleR::SingleR(test = query, ref = referenceData(), labels = referenceData()[[input$refLabels]])
            annotations_df = data.frame(annotations)
            annotations_df = annotations_df %>% dplyr::select(labels, pruned.labels)
            updatedSeuratObject = AddMetaData(updatedSeuratObject, metadata = annotations_df, col.name = c('SingleR_Celltypes', 'SingleR_pruned_Celltypes'))
          }
          
          timeEnd <- Sys.time()
          elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')

          showNotification(paste("Cell Type annotation Complete in", round(elapsedTime, 2), "mins."), type = "message")
          shinyalert("Cell Type Annotation Completed", type = "success", confirmButtonCol = "#1c82e0")
          
          seuratObj(updatedSeuratObject) # Update the main Seurat object 
          rm(updatedSeuratObject)
          rm(query)
          gc()
          
          # Reactive Reduction plot
          reductionplt <- reactive({
            if(("umap" %in% Reductions(seuratObj())) || ("tsne" %in% Reductions(seuratObj()))) {
              if(input$labelplot == FALSE) {
                p = DimPlot(seuratObj(), pt.size = input$ptSize, group.by = "SingleR_Celltypes", label = FALSE, label.size = 5) + 
                  RotatedAxis() + 
                  guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
                  theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), 
                        axis.title = element_text(colour = "black", size = 15, face = "bold"), 
                        legend.position = "right", legend.box = "vertical", 
                        panel.background = element_rect(colour = "white"))
                return(p)
              }
              if(input$labelplot == TRUE) {
                p = DimPlot(seuratObj(), pt.size = input$ptSize, group.by = "SingleR_Celltypes", label = TRUE, label.size = 5) + 
                  RotatedAxis() + 
                  guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
                  theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), 
                        axis.title = element_text(colour = "black", size = 15, face = "bold"), 
                        legend.position = "right", legend.box = "vertical", 
                        panel.background = element_rect(colour = "white"))
                return(p)
              }
            } 
          })
          
          # Render reduction plot
          output$redctionPlot <- renderPlotly({
            req(reductionplt())
            plot <- reductionplt()
            ggplotly(plot)
          })
          
          observe({
            if (!is.null(reductionplt())) {
              shinyjs::show("downloadPlot_CA")
              shinyjs::show("labelplot")
              shinyjs::show("ptSize")
            }
          })
          
          if (!is.null(reductionplt())) {
            shinyjs::show("annoplot")
          }
          
          observe({
            if((!is.null(annotations) == TRUE) || !is.null(reductionplt()))
            {
              shinyjs::show("downloadAnnotationResults")
            }
          })
          
          output$downloadPlot_CA <- downloadHandler(
            filename = function() {
              paste0("CogentDS_v2.0_CelltypeAnnotation_DimPlot", "-", Sys.Date(), input$annoplot)
            },
            content = function(file) {
              req(reductionplt())
              reductionplt()
              ggsave(file, plot = reductionplt(), width = 10, height = 10)
            },
            contentType = 'image'
          ) 
          
          # Download annotation results as a CSV file
          output$downloadAnnotationResults <- downloadHandler(
            filename = function() { paste("CogentDS_v2.0_CelltypeAnnotationResults-", Sys.Date(), ".csv", sep = "") },
            content = function(file) {
              req(annotations)
              write.csv(as.data.frame(annotations), file, row.names = TRUE)
            }
          )
        })
      }, error = function(e) {
        if (length(intersect(rownames(query), rownames(referenceData()))) == 0)
          shinyalert("Error", "No common rownames between query and reference", type = "error", confirmButtonCol = "#1c82e0")
      })
      
      shinyjs::enable("performCellAnnotations")
      
    }, ignoreNULL = FALSE)
  })
}
