#' UI for Clustering and Visualization Module
#'
#' Generates UI elements for a module dedicated to clustering and visualization within
#' a Shiny application. This module facilitates the execution of clustering algorithms
#' and the subsequent visualization of results through dimensionality reduction techniques
#' like UMAP and t-SNE. It provides users with action buttons to initiate clustering,
#' UMAP, and t-SNE computation and displays the visualizations in an interactive manner
#' using Plotly outputs.
#' 
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return A list of UI elements wrapped in a `tagList`, including action buttons
#'         for initiating clustering and dimensionality reduction, and Plotly output
#'         areas for displaying UMAP and t-SNE visualizations.
#' @export
#'
#' @examples
#' 
#' clusteringVisualizationUI("clusteringVisualizationModule")
clusteringVisualizationUI <- function(id) {
  ns <- NS(id)
  tagList(
    h2("Clustering & Non-linear dimension reduction"), 
    h5("*Required", style="color:#1c82e0"),
    fluidRow(
      column(4,
             numericInput(ns("dimsForClustering"), "Number of dimensions (PCs) for clustering", value = 10),
             bsPopover(ns("dimsForClustering"), "Number of dimensions", "Number of principal components to use for clustering", placement = "right", trigger = "hover", options = list(container = "body")),
             numericInput(ns("resolutionForClustering"), "Resolution for clustering", value = 0.8, step = 0.1),
             bsPopover(ns("resolutionForClustering"), "Resolution", "Value of the resolution parameter, use a value above (below) 1.0 if you want to obtain a larger (smaller) number of communities.", placement = "right", trigger = "hover", options = list(container = "body")),
             radioButtons(ns("slect_nondim"), "Select non-linear dimension reduction", choices = c("UMAP", "tSNE"), selected = NULL),
             bsPopover(ns("slect_nondim"), "Select non-linear dimension reduction approach. Options 1) UMAP: Uniform manifold approximation projection 2) tSNE: T-Stochastic neighbor embedding. Whatever option is chosen, is locked as the reduction for the rest of the analysis!", placement = "right", trigger = "hover", options = list(container = "body")),
             conditionalPanel(
               condition = "input.slect_nondim == 'UMAP'",
               numericInput(ns("umapDims"), "Number of dimensions (PCs) for UMAP", value = 10),
               bsPopover(ns("umapDims"), "Number of dimensions", "Number of principal components to use for UMAP reduction", placement = "right", trigger = "hover", options = list(container = "body")),
               numericInput(ns("nneighbors"), "Number of neighboring points for UMAP", value = 30),
               bsPopover(ns("nneighbors"), "Number of neighbors", "This determines the number of neighboring points used in local approximations of manifold structure. Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should often be in the range 5 to 50", placement = "right", trigger = "hover", options = list(container = "body")),
               numericInput(ns("mindist"), "Minimum distance for UMAP", value = 0.3),
               bsPopover(ns("mindist"), "Minimum distance", "This controls how tightly the embedding is allowed compress points together. Larger values ensure embedded points are more evenly distributed, while smaller values allow the algorithm to optimise more accurately with regard to local structure. Sensible values are in the range 0.001 to 0.5", placement = "right", trigger = "hover", options = list(container = "body")), ns = NS(id)),
             conditionalPanel(
               condition = "input.slect_nondim == 'tSNE'",
               numericInput(ns("tsneDims"), "Number of dimensions for tSNE", value = 10),
               bsPopover(ns("tsneDims"), "Number of dimensions", "Number of principal components to use for tSNE reduction", placement = "right", trigger = "hover", options = list(container = "body")),
               numericInput(ns("perplexity"), "Perplexity for tSNE", value = 30),
               bsPopover(ns("perplexity"), "Perplexity", "This determines the number of neighboring points used in local approximations of manifold structure. Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should often be in the range 5 to 50", placement = "right", trigger = "hover", options = list(container = "body")), ns = NS(id)),
             numericInput(ns("ptSize"), "Point size", value = 1),
             bsPopover(ns("ptSize"), "Point size", "Adjust point size", placement = "right", trigger = "hover", options = list(container = "body")),
             actionButton(ns("runClusteringanddimRed"), "Cluster cells & Perform non-linear reduction", class = "btn-primary", style='color:#fff; background-color:#1c82e0')),
      column(8,
             conditionalPanel(
               condition = "input.runClusteringanddimRed > 0",
               selectInput(ns("DimPlot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
               downloadButton(ns('downloadPlot'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
               plotlyOutput(ns("dimreductionPlot"), width = "700", height = "600"), ns = NS(id)))
    )
  )
}