#' Data Upload UI Component
#'
#' This function creates a set of UI elements for data uploading in a Shiny app. It includes 
#' a file input for dataset upload, a text input for specifying a project name, and numeric 
#' inputs for defining minimum cell and feature thresholds. The function is designed to be 
#' flexible, supporting both .csv and .rda file formats for dataset uploads.
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return A UI element comprising a file input, a project name text input, and numeric inputs 
#'         for minimum cells and features. These elements are wrapped in a `tagList` for easy 
#'         integration into the app's UI.
#' @export
#'
#' @examples
#' In the UI definition of a Shiny app
#' dataUploadUI("dataUploadModule")
#' 
#' @note The 'minCells' and 'minFeatures' inputs allow users to specify filtering criteria 
#'       for the dataset post-upload, offering a basic level of data preprocessing.
#'       
dataUploadUI <- function(id) {
  ns <- NS(id)
  
  tagList(
    fluidRow(column(4,
      h2('Upload Data for Step-by-Step Analysis'),
    radioButtons(ns("fileoption"), "Select type of data upload", choices = c("Processed Data from CogentAP", "Raw count matrix", "Example data"), selected = "Processed Data from CogentAP"),
    bsPopover(ns("fileoption"), "File upload options", "CogentDS accepts raw count matrices in .csv or .csv.gz formats and Processed Data from CogentAP in .rds or .rda formats. Please choose “Example data” to try all app functions in case you do not have your data in the required format. “Example data” consists of 2,000 PBMC cells and includes isoform data.", placement = "right", trigger = "hover", options = list(container = "body")),
    conditionalPanel(
      condition = "input.fileoption == 'Processed Data from CogentAP'",
      fileInput(ns("file"), "Upload your dataset", accept = c(".rda", ".rds")), ns = NS(id)),
    conditionalPanel(
      condition = "input.fileoption == 'Raw count matrix'",
      fileInput(ns("file_raw"), "Upload your dataset", accept = c(".gz", ".csv")), ns = NS(id)),
    conditionalPanel(
      condition = "input.fileoption == 'Processed Data from CogentAP' | input.fileoption == 'Raw count matrix'",
      textInput(ns("projectName"), "Project name", value = "Myproject"),
      bsPopover(ns("projectName"), "Project name", "Please set the name of the project to be used in Seurat Object", placement = "right", trigger = "hover", options = list(container = "body")),
      numericInput(ns("minCells"), "Minimum number of cells", value = 3),
      bsPopover(ns("minCells"), "Min. cells", "Include features detected in at least this many cells. Will subset the counts matrix as well. To reintroduce excluded features, create a new object with a lower cutoff", placement = "right", trigger = "hover", options = list(container = "body")),
      numericInput(ns("minFeatures"), "Minimum number of genes/features", value = 200),
      bsPopover(ns("minFeatures"), "Min. features", "Include cells where at least this many features are detected", placement = "right", trigger = "hover", options = list(container = "body")),
      selectInput(ns("metadata_question"), label = "Add metadata (Optional)", choices = c("Yes", "No"), selected = "No"),
      bsPopover(ns("metadata_question"), "Metadata", "Add Metadata, Please specify samples for each cell using SampleType column. SampleType is required", placement = "right", trigger = "hover", options = list(container = "body")),
      conditionalPanel(
        condition = "input.metadata_question == 'Yes'",
        fileInput(ns("metadata_file"), label = "Upload metadata file", accept = c(".csv")),
        selectInput(ns("samples_col"), "Select samples column", choices = NULL, selected = NULL, multiple = FALSE),
        selectizeInput(ns("samples_pick"), "Select samples to analyze", choices = NULL, selected = NULL, multiple = TRUE), ns=NS(id)), ns=NS(id)),
    conditionalPanel(
      condition = "input.fileoption == 'Raw count matrix'",
      fileInput(ns("geneInfofile"), h5(span("Upload gene Info file (Required)", style = "font-weight: bold"), span(icon("info-circle"), id = ns("icon3"), style = "color: white")), accept = c(".csv")),
      bsPopover(ns("icon3"), "Gene Info File", "CogentDS accepts gene info file which should have GeneID column which matches the first column in the counts matirx and GeneSymbol column corresponding to GeneID column.", placement = "right", trigger = "hover", options = list(container = "body")), ns=NS(id)),
    conditionalPanel(
      condition = "input.fileoption == 'Processed Data from CogentAP' | input.fileoption == 'Example data'",
      selectInput(ns("RNA_use"), "Select Gene Expression Profile for analysis", choices = c("Exon+Intron", "Exon-Only"), selected = "Exon+Intron", multiple = FALSE),
      bsPopover(ns("RNA_use"), "Gene Expression Profile", "CogentDS can analyze gene matrices with both Exon+Intron (where reads mapping to both exon and inron are counted) or Exon-Only.", placement = "right", trigger = "hover", options = list(container = "body")), ns=NS(id)),
    actionButton(ns("upload"), label = "Prepare Data for downstream analysis (Required)", class = "btn-primary", icon = icon("upload"), style='color:#fff; background-color:#1c82e0')),
    column(6,
      conditionalPanel(
        condition = "input.upload > 0",
        #h3("Data Stats"),
        selectInput(ns("readstats_select"), label = "Select Read Stats", choices = c("Read Stats", "Gene Body Assignment Breakdown", "Undesirable Read Breakdown", "Other Stats"), selected = "Read Stats"),
        DT::dataTableOutput(ns("seurat")), ns=NS(id))))
  )
}
