#' Dimension Reduction Server Logic
#' 
#' Handles server-side logic for dimension reduction operations on a Seurat object 
#' within a Shiny module.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'           
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#'
#' @seealso Refer to \code{\link[Seurat]{RunPCA}}, \code{\link[Seurat]{VizDimLoadings}}, 
#' \code{\link[Seurat]{DimHeatmap}}, \code{\link[Seurat]{ElbowPlot}} for more information 
#' 
#' @examples
#' dimReductionServer("dimReductionModule", seuratObject)
dimReductionServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    shinyjs::hide("downloadPlot_pcaplot")
    shinyjs::hide("downloadPlot_elbow")
    shinyjs::hide("downloadPlot_dimloadplot")
    shinyjs::hide("downloadPlot_dimheatplot")
    shinyjs::hide("pcaplot")
    shinyjs::hide("elbow")
    shinyjs::hide("dimloadplot")
    shinyjs::hide("dimheatplot")
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(seuratObj())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObj()@meta.data)
      
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObj()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    
    # Update selectInput choices for UMAP and t-SNE color by metadata
    observe({
      updateSelectInput(session, "PCAColorBy", choices = metadataColumns())
      bsPopover(ns("PCAColorBy"), "Select Cell Groups", "Name of one or more metadata columns to group (color) cells by (for example, orig.ident); pass 'ident' to group by identity class", placement = "right", trigger = "hover", options = list(container = "body"))
      
    })
    
    shiny::observeEvent(input$runPCAButton, {
      
      if (!"scale.data" %in% Layers(seuratObj())) {
        shinyalert("Error", "Please run Normalization module", type = "error", confirmButtonCol = "#1c82e0")
        return()
      }
      req(seuratObj())
      npcs <- input$npcs
      approx <- input$approx
      
      # Check if the specified number of pcs is less than or equal to the number of columns in the Seurat object
      if (npcs > ncol(seuratObj())) {
        #stop('number of pcs selected are more than cells')
        shinyjs::enable("runPCAButton")
        shinyalert("Error", "The number of Principal Components exceeds the available dimensions. Please reduce the number of Principal Components and try again.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      } else {
        shinyjs::disable("runPCAButton")
        # Start timing
        timeStart <- Sys.time()
        
        # Perform PCA with progress update
        withProgress(message = 'Performing PCA...', value = 0, {
          incProgress(0.5, detail = "Calculating PCA...")
          
          tryCatch({
            npcs <- input$npcs
            approx <- input$approx
            updatedSeuratObject = seuratObj()
            parameter_list = updatedSeuratObject@assays$user_parameter_tracker
            npcs_use = paste0('Number of PCs: ', npcs <- input$npcs)
            approximation = paste0('Approximation: ', input$approx)
            parameter_list = c(parameter_list, npcs_use, approximation)
            updatedSeuratObject@assays$user_parameter_tracker = NULL
            set.seed(12345)
            updatedSeuratObject <- Seurat::RunPCA(object = updatedSeuratObject, npcs = npcs, approx = approx)
            updatedSeuratObject@assays$user_parameter_tracker = parameter_list
            seuratObj(updatedSeuratObject) # Update the main Seurat object with new variable features
            rm(updatedSeuratObject)
            gc()
          }, error = function(e) {
            showNotification(paste("Error performing PCA:", e$message), type = "error")
            return(NULL)
          })
          
          incProgress(0.5, detail = "Finalizing...")
        })
        
        # End timing
        timeEnd <- Sys.time()
        elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
        
        showNotification(paste("PCA completed in", round(elapsedTime, 2), "mins."), type = "message")
        
        # Render PCA Scatter Plot
        output$PCAPlotOutput <- renderPlotly({
          req(seuratObj())
          if ("pca" %in% Reductions(seuratObj())){
          plot <- DimPlot(seuratObj(), group.by = input$PCAColorBy, reduction = "pca") + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          ggplotly(plot)
          } else NULL
        })
        
        pcaplt = reactive({
          DimPlot(seuratObj(), group.by = input$PCAColorBy, reduction = "pca") + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        })
        
        if(!is.null(pcaplt()))
        {
          shinyjs::show("downloadPlot_pcaplot")
          shinyjs::show("pcaplot")
        }
        
        output$downloadPlot_pcaplot <- downloadHandler(
          filename = function() {
            paste0("CogentDS_PCAPlot", "-", Sys.Date(), input$pcaplot)
          },
          content = function(file) {
            DimPlot(seuratObj(), group.by = input$PCAColorBy, reduction = "pca") + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            ggsave(file, plot = pcaplt(), width = 10, height = 10)
          },
          contentType = 'image'
        )
        
        # Render Elbow plot
        output$elbowPlot <- renderPlot({
          req(seuratObj())
          tryCatch({
            Seurat::ElbowPlot(seuratObj()) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white")) # Ensure ElbowPlot function returns a valid plot
          }, error = function(e) {
            showNotification(paste("Error rendering elbow plot:", e$message), type = "error")
            return(NULL)
          })
        })
        
        elbowplt = reactive({
          req(seuratObj())
          Seurat::ElbowPlot(seuratObj()) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white")) # Ensure ElbowPlot function returns a valid plot
        })
        
        if(!is.null(elbowplt()))
        {
          shinyjs::show("downloadPlot_elbow")
          shinyjs::show("elbow")
        }
        
        output$downloadPlot_elbow <- downloadHandler(
          filename = function() {
            paste0("CogentDS_ElbowPlot", "-", Sys.Date(), input$elbow)
          },
          content = function(file) {
            Seurat::ElbowPlot(seuratObj()) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            ggsave(file, plot = elbowplt(), width = 10, height = 10)
          },
          contentType = 'image'
        )
        
        # Render PCA loadings plot
        output$dimLoadingsPlot <- renderPlot({
          req(seuratObj())
          if (is.null(seuratObj())) {
            return(NULL)
          } else {
            Seurat::VizDimLoadings(seuratObj(), dims = 1:input$VizDims, reduction = "pca") & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          }
        })
        
        dimLoadingsplt = reactive({
          req(seuratObj())
          Seurat::VizDimLoadings(seuratObj(), dims = 1:input$VizDims, reduction = "pca") & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        })
        
        if(!is.null(dimLoadingsplt()))
        {
          shinyjs::show("downloadPlot_dimloadplot")
          shinyjs::show("dimloadplot")
        }
        
        output$downloadPlot_dimloadplot <- downloadHandler(
          filename = function() {
            paste0("CogentDS_DimLoadingsPlot", "-", Sys.Date(), input$dimloadplot)
          },
          content = function(file) {
            #Seurat::VizDimLoadings(seuratObj(), dims = 1:input$VizDims, reduction = "pca") & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            dimLoadingsplt()
            ggsave(file, plot = dimLoadingsplt(), width = 10, height = 10)
          },
          contentType = 'image'
        )
        
        # Render PCA heatmap plot
        output$dimHeatmapPlot <- renderPlot({
          req(seuratObj())
          if (is.null(seuratObj())) {
            return(NULL)
          } else {
            Seurat::DimHeatmap(seuratObj(), dims = 1:input$heatmapDims, cells = input$cells, balanced = TRUE, fast = F) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          }
        })
        
        dimHeatmapplt = reactive({
          req(seuratObj())
          Seurat::DimHeatmap(seuratObj(), dims = 1:input$heatmapDims, cells = input$cells, balanced = TRUE, fast = F) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        })
        
        if(!is.null(dimHeatmapplt()))
        {
          shinyjs::show("downloadPlot_dimheatplot")
          shinyjs::show("dimheatplot")
        }
        
        output$downloadPlot_dimheatplot <- downloadHandler(
          filename = function() {
            paste0("CogentDS_DimHeatmapPlot", "-", Sys.Date(), input$dimheatplot)
          },
          content = function(file) {
            Seurat::DimHeatmap(seuratObj(), dims = 1:input$heatmapDims, cells = input$cells, balanced = TRUE, fast= F) & theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 8, face = "bold"), axis.title = element_text(colour = "black", size = 8, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            ggsave(file, plot = dimHeatmapplt(), width = 10, height = 10)
          },
          contentType = 'image'
        )
        
        tryCatch({
          shinyalert("PCA Analysis is complete.", "Let's Get Started with Clustering", type = "success", confirmButtonCol = "#1c82e0")
        }, error = function(e) {
          shinyalert("Error Runnng PCA", "Please check input data", type = "error", confirmButtonCol = "#1c82e0")
        })
        shinyjs::enable("runPCAButton")
      }
    })
  })
}