#' Isoform analysis module server logic
#' 
#' This server module handles transcript data analysis and expression projection
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObj A reactive expression containing the Seurat object.
#'
#' @return
#' @export
#'
#' @examples
#' isoformServer("isoformModule", seuratObject)

isoformServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    data <- reactiveVal(NULL)
    isoform_data_seurat = reactiveVal(NULL)
    metadataColumns = reactiveVal(NULL)
    plotout = reactiveVal(NULL)
    isoform_DE_markers = reactiveVal(NULL)
    
    
    output$ConditionalParamteres <- renderUI({
      fluidRow(
        column(4,
               numericInput(ns("featureselection_transcript"), "Number of variable isoforms to select:", value = 2000),
               numericInput(ns("npcs_isoforms"), "Number of principal components", value = 50),
               numericInput(ns("dimsForClustering_isoforms"), "Number of dimensions for clustering", value = 10),
               numericInput(ns("resolutionForClustering_isoform"), "Resolution for clustering", value = 0.8),
               radioButtons(ns("isoform_de"), "Find isoform differential markers", choices = c(TRUE, FALSE), selected = FALSE),
               actionButton(ns("analysis_isoform_btn"), "Analyze isoform data", class = "btn-primary", style='color:#fff; background-color:#1c82e0')))
    })
    
    metadataColumns <- reactive({
      req(seuratObj())
      
      # Get the names of  metadata columns
      all_metadata <- colnames(seuratObj()@meta.data)
      
      # Filter out numeric columns
      relevant_columns <- all_metadata[sapply(seuratObj()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Return the filtered relevant columns
      return(relevant_columns)
    })
    
    isoform_data_seurat <- eventReactive(input$analysis_isoform_btn, {
      req(input$featureselection_transcript, seuratObj())
      shinyjs::disable("analysis_isoform_btn")
      withProgress(message = 'Performing Isoform Analysis:', value = 0, {
        incProgress(0.1, detail = "Checking Isoform Data")
        isoform_expression = reactive({
          req(seuratObj())
          parameter_list = seuratObj()@assays$user_parameter_tracker
          check_isoform_presence = grep('Isoform Data: ', parameter_list, value = T)
          if(length(check_isoform_presence) == 0)
          {
            return(NULL)
          }
          else
          {
            isoform_file_split = strsplit(check_isoform_presence, split = "Isoform Data: ", fixed = T)
            sc_iso_temp = readRDS(isoform_file_split[[1]][2])
            if(length(Cells(seuratObj())) > 30) {
              if(length(Cells(seuratObj())) > 10000) {
                shinyalert("Number of cells > 10,000", "Isoform analysis will be time & resource intensive", type = "warning", confirmButtonCol = "#1c82e0")
                return(sc_iso_temp)
              } else {
                return(sc_iso_temp)
              }
            } else {
              shinyalert("Not enough cells for Isoform Analysis", "Needs at-least > 30 cells", type = "warning", confirmButtonCol = "#1c82e0")
              return(NULL)
            } 
          }
        })
        incProgress(0.1, detail = "Isoform Data Check Compelete")
        if(is.null(isoform_expression()) == TRUE) {
          shinyalert("Isoform RNA Data Not Found", "Please check your input data! CogentAP output has both RNA and Transcript_RNA slots", type = "warning", confirmButtonCol = "#1c82e0")
          shinyjs::hide("featureselection_transcript")
          shinyjs::hide("npcs_isoforms")
          shinyjs::hide("dimsForClustering_isoforms")
          shinyjs::hide("resolutionForClustering_isoform")
          shinyjs::hide("isoform_de")
          shinyjs::hide("analysis_isoform_btn")
          isoform_seurat = NULL
          incProgress(0.8, detail = "Finished")
        } else {
          seuratObj_temp = seuratObj()
          parameter_list = seuratObj_temp@assays$user_parameter_tracker
          numvarfeatures = paste0('Isoform number of variable features: ', input$featureselection_transcript)
          npcs_isoforms_use = paste0('Isoform number of PCs: ', input$npcs_isoforms)
          dimensions = paste0('Isoform number of dimensions: ', input$dimsForClustering_isoforms)
          resolution_use = paste0('Isoform clustering resolution: ', input$resolutionForClustering_isoform)
          parameter_list = c(parameter_list, numvarfeatures, npcs_isoforms_use, dimensions, resolution_use)
          seuratObj_temp@assays$user_parameter_tracker = NULL
          incProgress(0.1, detail = "Isoform seurat object generated")
          isoform_seurat = isoform_expression()
          isoform_seurat = subset(isoform_seurat, cells = Cells(seuratObj_temp))
          rm(seuratObj_temp)
          incProgress(0.1, detail = "Normalizing")
          isoform_seurat = NormalizeData(isoform_seurat)
          incProgress(0.1, detail = "Finding Variable Features")
          isoform_seurat = FindVariableFeatures(isoform_seurat, nfeatures = input$featureselection_transcript)
          incProgress(0.1, detail = "Scaling")
          isoform_seurat = ScaleData(isoform_seurat)
          incProgress(0.1, detail = "PCA")
          isoform_seurat = RunPCA(isoform_seurat, npcs = input$npcs_isoforms)
          isoform_seurat = FindNeighbors(isoform_seurat, dims = 1:input$dimsForClustering_isoforms)
          incProgress(0.1, detail = "Finding Clusters")
          isoform_seurat = FindClusters(isoform_seurat, resolution = input$resolutionForClustering_isoform)
          incProgress(0.1, detail = "Non Linear Dimension Reduction & Markers")
          seuratObj_temp = seuratObj()
          seuratObj_temp@assays$user_parameter_tracker = NULL
          if("umap" %in% Reductions(seuratObj_temp)) {
            if(length(Cells(isoform_seurat)) < 30) {
              shinyalert("Number of cells found are less than number of neighbors required for UMAP", "Needs at-least 30 or more cells", type = "error", confirmButtonCol = "#1c82e0")
            } else {
              isoform_seurat = RunUMAP(isoform_seurat, dims = 1:input$dimsForClustering_isoforms)
              Idents(isoform_seurat) = "seurat_clusters"
              isoform_seurat[['gene_umap']] = CreateDimReducObject(embeddings = seuratObj_temp[['umap']]@cell.embeddings, assay = "RNA")
              seuratObj_temp[['isoform_umap']] = CreateDimReducObject(embeddings = isoform_seurat[['umap']]@cell.embeddings, assay = "RNA")
              if(input$isoform_de == TRUE) {
                if(length(Cells(isoform_seurat)) <= 5000) {
                  isoform_markers = FindAllMarkers(isoform_seurat, assay = "RNA")
                } else {
                  message('using max.cells.per.ident = 200')
                  isoform_markers = FindAllMarkers(isoform_seurat, assay = "RNA", max.cells.per.ident = 200)
                }
                isoform_markers = data.frame(isoform_markers)
                isoform_DE_markers(isoform_markers)
              }
              if(input$isoform_de == FALSE) {
                isoform_DE_markers(NULL)
              }
              ##### saving the processed isoform data 
              seuratObj_temp@assays$user_parameter_tracker = parameter_list
              seuratObj(seuratObj_temp)
              data(isoform_seurat)
            }
          }
          if("tsne" %in% Reductions(seuratObj_temp)) {
            if(length(Cells(isoform_seurat)) < 30) {
              shinyalert("Number of cells found are less than number of neighbors required for tSNE", "Needs at-least 30 or more cells", type = "error", confirmButtonCol = "#1c82e0")
            } else {
              isoform_seurat = RunTSNE(isoform_seurat, dims = 1:input$dimsForClustering_isoforms, check_duplicates = FALSE)
              Idents(isoform_seurat) = "seurat_clusters"
              isoform_seurat[['gene_tsne']] = CreateDimReducObject(embeddings = seuratObj_temp[['tsne']]@cell.embeddings, assay = "RNA")
              seuratObj_temp[['isoform_tsne']] = CreateDimReducObject(embeddings = isoform_seurat[['tsne']]@cell.embeddings, assay = "RNA")
              if(input$isoform_de == TRUE) {
                if(length(Cells(isoform_seurat)) <= 5000) {
                  isoform_markers = FindAllMarkers(isoform_seurat, assay = "RNA")
                } else {
                  message('using max.cells.per.ident = 200')
                  isoform_markers = FindAllMarkers(isoform_seurat, assay = "RNA", max.cells.per.ident = 200)
                }
                isoform_markers = data.frame(isoform_markers)
                isoform_DE_markers(isoform_markers)
              }
              if(input$isoform_de == FALSE) {
                isoform_DE_markers(NULL)
              }
              ##### saving the processed isoform data 
              seuratObj_temp@assays$user_parameter_tracker = parameter_list
              seuratObj(seuratObj_temp)
              data(isoform_seurat)
            }
          }
          incProgress(0.1, detail = "Finished")
          rm(isoform_seurat)
          rm(seuratObj_temp)
          gc()
        }
      })
      return(data())
    })
    
    observeEvent(!is.null(isoform_data_seurat()), {
      req(isoform_data_seurat())
      output$ConditionalParamteres <- renderUI({
        fluidRow(
          column(4,
                 radioButtons(ns("overlap_type_transcripts"), "Select Visualization:",
                              choices = c("Genes", "Isoforms", "Metadata", "Isoform Markers"),
                              selected = "Genes"),
                 conditionalPanel(
                   condition = "input.overlap_type_transcripts == 'Genes'",
                   selectizeInput(ns("isoform_genes_select"), "Select Genes", choices = NULL, selected = NULL, multiple = FALSE), ns = NS(id)),
                 conditionalPanel(
                   condition = "input.overlap_type_transcripts == 'Isoforms'",
                   selectizeInput(ns("isoform_isoform_select"), "Select Isoforms", choices = NULL, selected = NULL, multiple = FALSE), ns = NS(id)),
                 conditionalPanel(
                   condition = "input.overlap_type_transcripts == 'Metadata'", 
                   selectInput(ns("umapColorBy_transcripts"), "Color by:", choices = NULL, selected = NULL, multiple = FALSE), ns = NS(id)),
                 conditionalPanel(
                   condition = "input.overlap_type_transcripts == 'Isoform Markers'", 
                   radioButtons(ns("showisoformde_data"), "Select Isoforms Data / Visualization", choices = c("Heatmap", "DE Isoforms Table"), selected = "Heatmap"), ns = NS(id)),
                 actionButton(ns("overlay_btn"), "Generate Plots / Data", style='color:#fff; background-color:#1c82e0'),
                 br(),
                 downloadButton(ns('download_isoformobj'), 'Download Processed Isoform Seurat Object', style='color:#fff; background-color:#1c82e0')),
          column(8,
                 tabsetPanel(
                   tabPanel("Plots",
                            conditionalPanel(
                              condition = "input.overlap_type_transcripts == 'Genes' | input.overlap_type_transcripts == 'Isoforms' | input.overlap_type_transcripts == 'Metadata' | input.showisoformde_data == 'Heatmap'",
                              selectInput(ns("isoformPlt"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                              downloadButton(ns('downloadPlot_isoform'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                              shinycssloaders::withSpinner(plotOutput(ns("isoform_plot"), width = "700", height = "600"), size = 1), ns = NS(id))
                   ),
                   tabPanel("DE Table",
                            conditionalPanel(
                              condition = "input.showisoformde_data == 'DE Isoforms Table'",
                              downloadButton(ns('downloadPlot_isoform_data'), 'Save DE Isoforms Table', style='color:#fff; background-color:#1c82e0'),
                              shinycssloaders::withSpinner(DT::dataTableOutput(ns("isoform_de_table")), size = 1), ns = NS(id)))))
        )
      })
    })
    
    observe({
      req(input$overlap_type_transcripts)
      updateSelectizeInput(session, "isoform_genes_select", choices = Features(seuratObj()), server = T)
    })
    
    observe({ 
      req(input$overlap_type_transcripts)
      updateSelectizeInput(session, "isoform_isoform_select", choices = Features(isoform_data_seurat()), server = T)
    })
    
    observe({ 
      req(metadataColumns(), input$overlap_type_transcripts)
      updateSelectInput(session, "umapColorBy_transcripts", choices = metadataColumns())
    })
    
    plotout <- eventReactive(input$overlay_btn, {
      req(seuratObj(), input$overlap_type_transcripts, isoform_data_seurat())
      shinyjs::disable("isoform_de")
      tryCatch({
        data_genes = seuratObj()
        data_isoforms = isoform_data_seurat()
        if(input$overlap_type_transcripts == "Genes")
        {
          req(input$isoform_genes_select) 
          if("isoform_umap" %in% Reductions(data_genes)) {
            p1 = FeaturePlot(data_genes, features = input$isoform_genes_select, keep.scale = "all", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("umap_1 based on gene exp") & ylab('umap_2 based on gene exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            p2 = FeaturePlot(data_genes, features = input$isoform_genes_select, keep.scale = "all", reduction = "isoform_umap", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("umap_1 based on isoform exp") & ylab('umap_2 based on isoform exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            #plot = ggpubr::ggarrange(plotlist = list(p1,p2))
            plot = p1+p2
            shinyjs::enable("isoform_de")
            return(plot)
          }
          if("isoform_tsne" %in% Reductions(data_genes)) {
            p1 = FeaturePlot(data_genes, features = input$isoform_genes_select, keep.scale = "all", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("tsne_1 based on gene exp") & ylab('tsne_2 based on gene exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            p2 = FeaturePlot(data_genes, features = input$isoform_genes_select, keep.scale = "all", reduction = "isoform_tsne", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("tsne_1 based on isoform exp") & ylab('tsne_2 based on isoform exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            #plot = ggpubr::ggarrange(plotlist = list(p1,p2), ncol = 2, nrow = 1, align = "hv")
            plot = p1+p2
            shinyjs::enable("isoform_de")
            return(plot)
          }
        }
        if(input$overlap_type_transcripts == "Isoforms")
        {
          req(input$isoform_isoform_select)
          if("gene_umap" %in% Reductions(data_isoforms)) {
            p1 = FeaturePlot(data_isoforms, features = input$isoform_isoform_select, keep.scale = "all", reduction = "gene_umap", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("umap_1 based on gene exp") & ylab('umap_2 based on gene exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            p2 = FeaturePlot(data_isoforms, features = input$isoform_isoform_select, keep.scale = "all", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("umap_1 based on isoform exp") & ylab('umap_2 based on isoform exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          }
          if("gene_tsne" %in% Reductions(data_isoforms)) {
            p1 = FeaturePlot(data_isoforms, features = input$isoform_isoform_select, keep.scale = "all", reduction = "gene_tsne", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("tsne_1 based on gene exp") & ylab('tsne_2 based on gene exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            p2 = FeaturePlot(data_isoforms, features = input$isoform_isoform_select, keep.scale = "all", pt.size = 1) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & xlab("tsne_1 based on isoform exp") & ylab('tsne_2 based on isoform exp') & theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          }
          #plot = ggpubr::ggarrange(plotlist = list(p1,p2))
          plot = p1+p2
          shinyjs::enable("isoform_de")
          return(plot)
        }
        if(input$overlap_type_transcripts == "Metadata")
        {
          req(input$umapColorBy_transcripts)
          data_isoforms = AddMetaData(data_isoforms, metadata = data_genes[[input$umapColorBy_transcripts]], col.name = "TransferredLabels")
          p1 = DimPlot(data_genes, pt.size = 1, group.by = input$umapColorBy_transcripts, label = F, label.size = 5) + RotatedAxis() + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(fill = "white")) + ggtitle('Gene Exp Based UMAP/tSNE Plot')
          p2 = DimPlot(data_isoforms, pt.size = 1, group.by = "TransferredLabels", label = F, label.size = 5) + RotatedAxis() + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(fill = "white")) + ggtitle('Isoform Exp Based UMAP/tSNE Plot')
          #plot = ggpubr::ggarrange(plotlist = list(p1,p2))
          plot = p1+p2
          shinyjs::enable("isoform_de")
          return(plot)
        }
        if(input$showisoformde_data == "Heatmap")
        {
          req(isoform_DE_markers())
          markers_temp = isoform_DE_markers()
          markers_temp %>% group_by(cluster) %>% dplyr::filter(avg_log2FC > 1) %>% slice_head(n = 5) %>% ungroup() -> top5
          plot = DoHeatmap(isoform_data_seurat(), features = top5$gene) + NoLegend()
          return(plot)
        }
        shinyalert("Plot is successfully generated", "Please try other plotting options!", type = "success", confirmButtonCol = "#1c82e0")
      }, error = function(e) {
        shinyalert("Plot failed", "Please check input data", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    
    output$isoform_de_table <- DT::renderDataTable({
      req(isoform_DE_markers())
      markers_temp = isoform_DE_markers()
      markers_temp = markers_temp %>% dplyr::filter(p_val_adj <= 0.05)
      DT::datatable(markers_temp)
    })
    
    output$isoform_plot <- renderPlot({
      req(plotout())
      plotout()
    })
    
    output$downloadPlot_isoform <- downloadHandler(
      filename = function() {
        paste0("CogentDS_IsoformAnalysis", "-", Sys.Date(), input$isoformPlt)
      },
      content = function(file) {
        req(plotout())
        plotout()
        ggsave(file, plot = plotout(), width = 10, height = 10)
      },contentType = 'image'
    )
    
    output$downloadPlot_isoform_data <- downloadHandler(
      filename = function() {
        paste0("CogentDS_DE_IsoformTable", "-", Sys.Date(), ".csv")
      },
      content = function(file) {
        req(isoform_DE_markers())
        markers_temp = isoform_DE_markers()
        markers_temp = markers_temp %>% dplyr::filter(p_val_adj <= 0.05)
        write.csv(markers_temp, file, sep=",", quote=F)
      }
    )
    
    output$download_isoformobj <- downloadHandler(
      filename = function() {
        paste0("CogentDS_Analysis_Processed_Isoform_Seurat_v5_Object", "-", Sys.Date(), ".rds")
      },
      content = function(file) {
        withProgress(message = 'Preparing Data Download...', value = 0, {
          incProgress(0.5, detail = "Downloading...")
          tryCatch({
            req(isoform_data_seurat())
            saveRDS(isoform_data_seurat(), file = file)
            shinyalert('Processed Isoform Seurat v5 Object Downloaded', "Let's Try Gene Fusion Analysis", type = "success", confirmButtonCol = "#1c82e0")
          }, error = function(e) {
            shinyalert('Error in downloading the processed Isoform Seurat v5 Object', 'Please check your input parameters', type = "error", confirmButtonCol = "#1c82e0")
          })
          incProgress(0.5, detail = "Downloaded")
        })
      }
    )
    return(data)
  })
}
