#' Normalization and Scaling Server Logic
#' 
#' This function handles server-side logic for normalization and scaling operations
#' on a Seurat object within a Shiny application. Users are prompted for parameters
#' through modal dialogs, allowing for a flexible and interactive normalization and
#' scaling process tailored to their data. The function supports various normalization
#' methods and applies scaling to prepare the data for downstream analysis.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObject A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' 
#' @importFrom shiny moduleServer observeEvent showModal modalDialog showNotification
#' @importFrom Seurat NormalizeData ScaleData
#'
#' @export
#'
#' @seealso \code{\link[Seurat]{NormalizeData}}, \code{\link[Seurat]{ScaleData}}
#'          for detailed explanations on normalization and scaling methods available in Seurat.
#'
#' @examples
#' normalizationScalingServer("normalizationScalingModule", seuratObject)
normalizationScalingServer <- function(id, seuratObject) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    shinyjs::hide("downloadPlot_varfeature")
    shinyjs::hide("varfeature")
  
    # Perform normalization using user-defined parameters
    shiny::observeEvent(input$performNormalization, {
      shinyjs::disable("performNormalization")
      req(seuratObject())
    
      # Perform normalization with selected method and scale factor
      if(input$normMethod == "LogNormalize") {
        withProgress(
          message = 'Performing Normalization with LogNormalize...',
          value = 0, {
            tryCatch({
              # Start timing
              timeStart <- Sys.time()
              incProgress(0.5)
              gc()
              normalizedSeuratObject = seuratObject()
              DefaultAssay(normalizedSeuratObject) <- 'RNA'
              parameter_list = normalizedSeuratObject@assays$user_parameter_tracker
              normalizemethod = paste0('Normalization Method: ', input$normMethod)
              scalefactor = paste0('Scale Factor: ', input$scaleFactor)
              featureselection_method = paste0('Feature Selection Method: ', input$featureselection)
              numberofvarfeatures = paste0('Number of Variable Features: ', input$featureselection_number)
              parameter_list = c(parameter_list, normalizemethod, scalefactor, featureselection_method, numberofvarfeatures)
              normalizedSeuratObject@assays$user_parameter_tracker = NULL
              if("pca" %in% Reductions(normalizedSeuratObject)) {
                normalizedSeuratObject@reductions$pca = NULL
              }
              if("umap" %in% Reductions(normalizedSeuratObject)) {
                normalizedSeuratObject@reductions$umap = NULL
              }
              normalizedSeuratObject <- Seurat::NormalizeData(object = normalizedSeuratObject, 
                                                            normalization.method = input$normMethod, 
                                                            scale.factor = input$scaleFactor)
              normalizedSeuratObject = FindVariableFeatures(normalizedSeuratObject, selection.method = input$featureselection, nfeatures = input$featureselection_number)
              if(input$usefeatures == "Variable features (Recommended for Large Datasets)")
              {
                normalizedSeuratObject = ScaleData(normalizedSeuratObject, features = VariableFeatures(normalizedSeuratObject))
              }
              if(input$usefeatures == "All features")
              {
                normalizedSeuratObject = ScaleData(normalizedSeuratObject, features = rownames(normalizedSeuratObject))
              }

              output$Varfeatures = renderPlot({
                top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                # plot variable features with and without labels
                plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
              })

              varfeatureplt = reactive({
                req(seuratObject())
                top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                # plot variable features with and without labels
                plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
              })

              if(!is.null(varfeatureplt()))
              {
                shinyjs::show("downloadPlot_varfeature")
                shinyjs::show("varfeature")
                
              }

              output$downloadPlot_varfeature <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_Top10_VariableFeature", "-", Sys.Date(), input$varfeature)
                },
                content = function(file) {
                   top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                   plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                # plot variable features with and without labels
                   LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
                   ggsave(file, plot = varfeatureplt(), width = 10, height = 10)
                },
                contentType = "image"
              )

              # Update the reactive Seurat object
              normalizedSeuratObject@assays$user_parameter_tracker = parameter_list
              seuratObject(normalizedSeuratObject)
              gc()
              # Notify user of completion and show elapsed time
              timeEnd <- Sys.time()
              elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
              shinyalert("Normalization, Feature Selection and Scaling completed.", "Let's Get Started with Linear Dimension Reduction", type = "success", confirmButtonCol = "#1c82e0")
              # Close the modal dialog after normalization
            }, error = function(e) {
              shinyalert("Normalization, Feature Selection and Scaling Failed.", "Please check input parameters", type = "error", confirmButtonCol = "#1c82e0")
            })
          }   
        )
      }  
      if(input$normMethod == "SCT") {
        withProgress(    
          message = 'Performing Normalization with SCT...',
          value = 0, {
            tryCatch({
              # Start timing
              timeStart <- Sys.time()
              incProgress(0.5)
              gc()
              normalizedSeuratObject = seuratObject()
              parameter_list = normalizedSeuratObject@assays$user_parameter_tracker
              normalizemethod = paste0('Normalization Method: ', input$normMethod)
              numberofvarfeatures = paste0('Number of Variable Features: ', input$featureselection_number)
              parameter_list = c(parameter_list, normalizemethod, numberofvarfeatures)
              normalizedSeuratObject@assays$user_parameter_tracker = NULL
              if("pca" %in% Reductions(normalizedSeuratObject)) {
                normalizedSeuratObject@reductions$pca = NULL
              }
              if("umap" %in% Reductions(normalizedSeuratObject)) {
                normalizedSeuratObject@reductions$umap = NULL
              }
              normalizedSeuratObject <- SCTransform(normalizedSeuratObject, vars.to.regress = "percent.mt", variable.features.n = input$featureselection_number, verbose = FALSE)

              output$Varfeatures = renderPlot({
                top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                # plot variable features with and without labels
                plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
              })

              varfeatureplt = reactive({
                req(seuratObject())
                top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                # plot variable features with and without labels
                plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
              })

              if(!is.null(varfeatureplt()))
              {
                shinyjs::show("downloadPlot_varfeature")
                shinyjs::show("varfeature")
              }

              output$downloadPlot_varfeature <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_Top10_VariableFeature", "-", Sys.Date(), input$varfeature)
                },
                content = function(file) {
                   top10 <- head(VariableFeatures(normalizedSeuratObject), 10)
                # plot variable features with and without labels
                  plot1 <- VariableFeaturePlot(normalizedSeuratObject)
                  LabelPoints(plot = plot1, points = top10, repel = TRUE) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
                  ggsave(file, plot = varfeatureplt(), width = 10, height = 10)
                },
                contentType = "image"
              )
              # Update the reactive Seurat object
              normalizedSeuratObject@assays$user_parameter_tracker = parameter_list
              seuratObject(normalizedSeuratObject)
              gc()
              # Notify user of completion and show elapsed time
              timeEnd <- Sys.time()
              elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
              shinyalert("Normalization, Feature Selection and Scaling completed.", "Let's Get Started with Linear Dimension Reduction", type = "success", confirmButtonCol = "#1c82e0")
            }, error = function(e) {
              shinyalert("Normalization, Feature Selection and Scaling Failed.", "Please check input parameters", type = "error", confirmButtonCol = "#1c82e0")
            })
          }
        )
      }
      shinyjs::enable("performNormalization")
    })
  })
}