#' Server logic for RNA Biotype Abundance Module
#' 
#' Implements server-side functionality for RNA abundance
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#' 
#' @seealso \code{\link[Seurat]{FindClusters}}, \code{\link[Seurat]{RunUMAP}}, and
#'          \code{\link[Seurat]{RunTSNE}} for more details on the underlying analysis
#'          methods used.
#'
#' @examples
#' rnaBiotypeServer("RNABiotypeModule", seuratObject_decontX)

rnaBiotypeServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    abundance_plot = reactiveVal(NULL)

    abundance_plot <- reactive({
    	req(seuratObj())
    	seuratObj_use = seuratObj()
      if(("Gene_Biotype" %in% colnames(seuratObj_use[['RNA']][[]])) && ("Gene_Length" %in% colnames(seuratObj_use[['RNA']][[]]))) {
        withProgress(message = 'Generating RNA-Biotypes Abundance Plot:', value = 0, {
          incProgress(0.3, detail = "Calculating Average Abundance across RNA-Biotypes")
          avg_expression = AggregateExpression(seuratObj_use, assays = "RNA", group.by = "orig.ident")
    	    avg_expression = data.frame(avg_expression)
    	    avg_expression$RNA_Biotype = seuratObj_use[['RNA']][[]]$Gene_Biotype
          incProgress(0.5, detail = "Preparing data for plotting")
    	    avg_expression_plot = reshape2::melt(avg_expression, id.vars = "RNA_Biotype", value.name = "RNA")
          plot = ggplot(avg_expression_plot, aes(x=log2(RNA+1), y=reorder(RNA_Biotype, -log2(RNA+1)))) + geom_boxplot(alpha = 0.5) + theme(axis.text = element_text(colour = "black", size = 10, face = "bold"), axis.title = element_text(colour = "black", size = 10, face = "bold"), legend.position = "none", panel.background = element_rect(colour = "white")) + xlab("Average abundance [Log2]") + ylab("RNA Biotypes")
          if(length(unique(avg_expression_plot$variable)) > 1) {
            avg_expression_plot$variable = gsub('RNA.', '', avg_expression_plot$variable)
            plot = plot + facet_wrap(~avg_expression_plot$variable, ncol = 3)
          }
          incProgress(0.1, detail = "Plot generated")
          rm(avg_expression)
        })
        shinyalert("RNA Biotypes Plot Generated", "Next Step: Let's perform Ambient RNA correction", type = "success", confirmButtonCol = "#1c82e0")
    	  return(plot)
       } else {
        shinyalert("RNA Biotypes not found in the processed data", "Please see: CogentAP v3.1 processed data contains rna biotype information", type = "warning", confirmButtonCol = "#1c82e0")
        return(NULL)
       }
    })

    output$rnaBiotypeAbundancePlot <- renderPlot({
      req(abundance_plot())
      if(!is.null(abundance_plot())) {
        abundance_plot()
      }
    })

    output$downloadPlot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_RNA-Biotypes_AbundancePlot", "-", Sys.Date(), input$abundanceplot)
      },
      content = function(file) {
        req(abundance_plot())
        if(!is.null(abundance_plot())) {
          abundance_plot()
          ggsave(file, plot = abundance_plot(), width = 10, height = 10)
        } else {
          shinyalert("RNA Biotypes not found in the processed data", "Please see: CogentAP v3.1 processed data contains rna biotype information", type = "warning", confirmButtonCol = "#1c82e0")
        }
      },
      contentType = 'image'
    )
  })
}
